#include <ctype.h>
#include <grp.h>
#include <iconv.h>
#include <locale.h>
#include <ncursesw/curses.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <wchar.h>
#include <wctype.h>
#include "agi-namelookup.h"
#include "agiconf.h"
#include "confvers.h"


/*
*    Configuration program for namelookup.agi
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/

/*
*    Disable these warnings
*/
#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wstringop-truncation"


/*  Debug  */
int debflg;

/*  Text output  */
int txtflg;

/*  UTF-8  */
int utflg;

/*  ASCII scroll progress bar  */
int ascflg;

/*  Group name  */
char grpnam[32];

/*  Home directory  */
char *home;

/*  Screen size  */
int cols, rows;

/*  Color on / off  */
int colflg;

/*  Phone mumber table  */
struct phnum *nums;
int numcnt, numentr, numax, numsiz;

/*  Country  */
int ctrycnt;
struct ctry country;
struct ctry countries[ANL_COUNTRIES_SIZ];

/*  Do we want web lookups  */
int webconf;

/*  Do we want compact files  */
int cmpconf;

/*  Convblockl done  */
int blkconf;

/*  Convnums done  */
int numconf;

/*  Area config  */
int areacnt, arconf;
struct area areas[ANL_AREAS_SIZ];

/*  Region config  */
int regconf;

/*  Mouse events  */
MEVENT msevent;


/*  Help  */
void hlp(void)
{
    fprintf(stderr, "Configure databases for namelookup.agi\n");
    fprintf(stderr, \
      "Usage: agiconf [-a] [-c] [-d] [-g Group] [-h] [-t] [-v]\n\n");
    fprintf(stderr, "-a: ASCII: Use space in scroll progress bar.\n");
    fprintf(stderr, "-c: Use color; Blue background.\n");
    fprintf(stderr, "-d: Debug on.\n");
    fprintf(stderr, "-g: Set group name. Default: asterisk\n");
    fprintf(stderr, "-h: Print this Help and exit.\n");
    fprintf(stderr, "-t: Output tsv files.\n");
    fprintf(stderr, "-v: Print version and exit.\n\n");
}

void prtvers(void)
{
    printf("Version: %s\n", version);
    #ifdef ANL_WITH_WIDE_NAMES
     printf("With name length %d bytes\n", ANL_NAME_MAX);
    #endif
    printf("\n");
}

/*
*    Screen functions
*/

/*
*  Process mouse events 
*  Returns Y position on left click and left double click
*/
int procmouse(void)
{
    if (getmouse(&msevent) == OK) {
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: procmouse(): Mouse event: Y: %d X: %d \n", \
	      msevent.y, msevent.x);
	}
	if (msevent.bstate & (BUTTON1_CLICKED | BUTTON1_DOUBLE_CLICKED)) {
	    if (debflg != 0)
		fprintf(stderr, "Debug: procmouse(): Button 1 ckicked\n");
	    if (msevent.x < cols - 2)
		return(msevent.y);
	}
    }

    /* There are no menu items at 0  */
    return(0);
}

/*  Process resize  */
void procresiz(void)
{
    move(22, 78);
    printw(" ");
    getmaxyx(stdscr, rows, cols);
    if(debflg != 0)
	fprintf(stderr, "Debug: procresiz(): Resolution: %dx%d\n", cols, rows);
}

/*  Clear a line  */
void clrline(int y, int x, int max)
{
    int i;

    move(y, x);

    i = x;
    while (i < max) {
	printw(" ");
	i++;
    }
}

/*  Draw a menu  */
void drwmen(struct menit *items, int y, int x, int mode)
{
    int i, j;

    i = 0;	/*  Vertical position  */
    j = 1;	/*  Menu item  */

    erase();
    box(stdscr, 0, 0);
    attrset(A_NORMAL);
    move(y, x);

    while (items[i].id != NULL && items[i].name != NULL && \
      items[i].id[0] != 0 && items[i].name[0] != 0) {
	move(y + i, x);
	if (items[i].name[0] == ' ') {
	    /*  A blank line  */
	    mvhline(y + i, 1, 0, cols - 2);
	    mvaddch(y + i, 0, ACS_LTEE);
	    mvaddch(y + i, cols - 1, ACS_RTEE);
	} else {
	    /*  Not a blank line  */
	    if (items[i].id[0] != ' ') {
		/*  Menu option  */
		attron(A_BOLD | A_UNDERLINE);
		printw("%s", items[i].id);
		attrset(A_NORMAL);
		move(y + i, x + 2);
		if (j == mode) {
		    if(colflg == 0)
			attron(A_REVERSE);
		    else
			attron(COLOR_PAIR(selctd));
		}
		printw("%s", items[i].name);
		if (j == mode) {
		    if(colflg == 0)
			attrset(A_NORMAL);
		    else
			attroff(COLOR_PAIR(selctd));
		}
		j++;
	    } else {
		/*  Not a menu option  */
		if (i == 0 && colflg != 0) {
		    /*  Header line  */
		    attron(A_BOLD);
		    attron(COLOR_PAIR(headtxt));
		}
		printw("%s", items[i].name);
		if (i == 0 && colflg != 0) {
		    attroff(COLOR_PAIR(headtxt));
		    attrset(A_NORMAL);
		}
	    }
	}
	i++;
    }

    attrset(A_NORMAL);
}

/*  Draw 'progress' bar  */
void drwbar(int y, int x, int curr, int cnt, int max)
{
    int pos, frac;

    pos = 8 * curr * max / cnt;
    frac = pos % 8;
    pos = pos / 8;

    move(y + pos, x);
    if (utflg == 0 || ascflg != 0) {
	/*  Reverse video space  */
	attron(A_REVERSE);
	printw(" ");
	attrset(A_NORMAL);
    } else {
	/*  UTF-8; Use block drawing chars 0x2581 - 0x2588  */
	if (frac == 0) {
	    printw("%lc", 0x2588);
	} else {
	    printw("%lc", 0x2588 - frac);
	    move(y + pos + 1, x);
	    attron(A_REVERSE);
	    printw("%lc", 0x2588 - frac);
	    attrset(A_NORMAL);
	}
    }
}

/*  Draw country selection  */
void drwctry(struct ctry *countries, int offs, int curr, int max, int mode)
{
    int i, ptr;

    i = 0;
    ptr = 0;

    move(1, 1);
    if (colflg != 0) {
	attron(A_BOLD);
	attron(COLOR_PAIR(headtxt));
    }
    printw("Select country");
    if (colflg != 0) {
	attroff(COLOR_PAIR(headtxt));
	attrset(A_NORMAL);
    }
    clrline(1, 25, cols - 1);
    if(colflg != 0)
	attron(COLOR_PAIR(highlt));
    if (country.num[0] != 0) {
	move(1, 25);
	printw("%s", country.num);
    }
    if (country.name[0] != 0) {
	move(1, 30);
	printw("%ls", country.name);
    }
    if(colflg != 0)
	attroff(COLOR_PAIR(highlt));
    move(3, 1);
    if (colflg != 0) {
	attron(A_BOLD);
	attron(COLOR_PAIR(headtxt));
    }
    printw("Country");
    move(3, 50);
    printw("Code");
    move(3, 60);
    printw("Int Pfx");
    move(3, 70);
    printw("Trunk Pfx");
    if(cols > 84) {
	move(3, cols - 2);
	printw("%%");
    }
    if (colflg != 0) {
	attroff(COLOR_PAIR(headtxt));
	attrset(A_NORMAL);
    }
    
    move(5, 1);

    i = 0;
    while (i <= max) {
	ptr = i + offs;
	if(countries[ptr].name[0] == 0)
	    break;
	move(5 + i, 1);
	if (ptr == curr) {
	    if(colflg != 0 && mode == 0)
		attron(COLOR_PAIR(selctd));
	    else
		attron(A_REVERSE);
	    //clrline(5 + i, 1, cols - 1);
	    clrline(5 + i, 1, cols - 3);
	    move(5 + i, 1);
	}
	printw("%ls", countries[ptr].name);
	move(5 + i, 50);
	printw("%s", countries[ptr].num);
	move(5 + i, 60);
	if(countries[ptr].incpfx [0] != 0)
	    printw("%s", countries[ptr].incpfx);
	move(5 + i, 70);
	if(countries[ptr].trkpfx [0] != 0)
	    printw("%s", countries[ptr].trkpfx);
	if (ptr == curr) {
	    if(colflg != 0 && mode == 0)
		attroff(COLOR_PAIR(selctd));
	    else
		attrset(A_NORMAL);
	}
	i++;
    }
}

/*  Draw areas selection  */
void drwareas(struct area *areas, int offs, int curr, int max, int mode)
{
    int i, ptr;

    i = 0;
    ptr = 0;

    move(1, 1);
    if (colflg != 0) {
	attron(A_BOLD);
	attron(COLOR_PAIR(headtxt));
    }
    printw("Select Areas");
    move(3, 1);
    printw("Country");
    move(3, 40);
    printw("Language");
    move(3, 60);
    printw("Resolution");
    move(3, cols - 2);
    printw("%%");
    if (colflg != 0) {
	attroff(COLOR_PAIR(headtxt));
	attrset(A_NORMAL);
    }

    while(i <= max) {
	ptr = i + offs;
	if(areas[ptr].iso[0] == 0)
	    break;
	move(5 + i, 1);
	if (areas[ptr].sel == 0) {
	    printw("[ ]");
	} else {
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	}
	move(5 + i, 5);
	if (ptr == curr) {
	    if(colflg != 0 && mode == 0)
		attron(COLOR_PAIR(selctd));
	    else
		attron(A_REVERSE);
	    clrline(5 + i, 5, cols - 3);
	    move(5 + i, 5);
	}
	printw("%ls", areas[ptr].country);
	move(5 + i, 40);
	if(areas[ptr].lang[0] != 0)
	    printw("%ls", areas[ptr].lang);
	move(5 + i, 60);
	if(areas[ptr].res[0] != 0)
	    printw("%ls", areas[ptr].res);
	if (ptr == curr) {
	    if(colflg != 0 && mode == 0)
		attroff(COLOR_PAIR(selctd));
	    else
		attrset(A_NORMAL);
	}
	i++;
    }
}

/*  Quit  */
void quit(int mode)
{
    if (nums != NULL) {
	free(nums);
	nums = NULL;
    }

    if (mode != 0) {
	/*  Switch off Ncurses mode  */
	clear();
	move(23, 0);
	refresh();
	endwin();
    }
}


/*
*    Data functions
*/

/*  Allocate memory for numbers  */
int initnums(void)
{
    int size;

    /*
    *  This thing only grows.
    *  numcnt may be reset though.
    */

    /*  Allocate  */
    if (nums == NULL) {
	/*  First time  */
	numentr = ANL_DFLT_SIZ;			/*  Nr of entries available  */
	numax = numentr - 2;			/*  Maximum  */
	size = sizeof(struct phnum) * numentr;	/*  Size in bytes  */
	numsiz = size;
	nums = malloc((size_t) size);
	if (nums == NULL) {
	    quit(0);	/*  Not yet in ncurses  */
	    fprintf(stderr, "initnums(): Can't allocate memory for numbers\n");
	    exit(1);
	}
	memset(nums, 0, size);
    } else {
	/*  Too little, double the size  */
	size = 2 * sizeof(struct phnum) * numentr;
	nums = realloc(nums, (size_t) size);
	if (nums == NULL) {
	    quit(1);	/*  In ncurses  */
	    fprintf(stderr, "initnums(): Can't allocate memory for numbers\n");
	    exit(1);
	}
	memset(nums + numentr, 0, numsiz);
	/*  Set new values  */
	numentr = numentr * 2;
	numax = numentr - 2;
	numsiz = size;
    }

    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: initnums(): Allocated %d bytes for %d numbers at %lx\n", \
	  size, numentr, (long unsigned int) nums);
	fprintf(stderr, "Debug: initnums(): Max: %d\n", numax);
    }

    return(0);
}

/*
*  Convert UTF-8 string to wide chracter string.
*  This function is used with non UTF-8 locales.
*  Note: iconv() wants char, but we need wchar_t.
*/
int icutftowc(char *inbuf, int inlen, wchar_t *outbuf, int outsiz)
{
    char *ip, *op, *le;
    int isle, outlen;
    size_t inbytesleft, outbytesleft;
    iconv_t cd;

    outlen = 0;

    outsiz = outsiz * sizeof(wchar_t);	/*  From wchar_t to char  */
    inbytesleft  = (size_t) inlen;
    outbytesleft = (size_t) outsiz;
    ip = inbuf;
    op = (char *) outbuf;	    	/*  From wchar_t to char  */

    /*  Endian test  */
    isle = 1;
    le = (char *) &isle;
    if (*le == 0) {
	/* Is big endian  */
	isle = 0;
    }

    if (isle == 0) {
	/*  Big endian  */
	if (sizeof(wchar_t) == 2)
	    cd = iconv_open("UCS-2BE//", "UTF-8//");
	else if (sizeof(wchar_t) == 4)
	    cd = iconv_open("UCS-4BE//", "UTF-8//");
	else
	    return(-1);
    } else {
	/*  Little endian  */
	if (sizeof(wchar_t) == 2)
	    cd = iconv_open("UCS-2LE//", "UTF-8//");
	else if (sizeof(wchar_t) == 4)
	    cd = iconv_open("UCS-4LE//", "UTF-8//");
	else
	    return(-1);
    }
    if (cd < 0) {
	fprintf(stderr, "icutftowc(): iconv_open failed\n");
	return(-1);
    }

    if (iconv(cd, &ip, &inbytesleft, &op, &outbytesleft) < 0) {
	fprintf(stderr, "icutftowc(): iconv failed\n");
	iconv_close(cd);
	return(-1);
    }
    iconv_close(cd);
    outlen = outsiz - ((int) outbytesleft);
    outlen = outlen / sizeof(wchar_t);	/*  From char to wchar_t  */
    outbuf[outlen] = 0;			/*  Terminate  */
    if (debflg != 0) {
	fprintf(stderr, "Debug: icutftowc(): converted to len: %d\n", \
	  outlen);
    }

    return(outlen);
}

/*  Decide if we can use mbstowcs()  */
int utfstowcs(char *inbuf, int inlen, wchar_t *outbuf, int outsiz)
{
    int ret;

    ret = 0;

    if (utflg == 0) {
	/*  Legacy locale  */
	ret = icutftowc(inbuf, inlen, outbuf, outsiz);
    } else {
	/*  UTF-8 locale  */
	if (mbstowcs(outbuf, inbuf, outsiz) < 0) {
	    fprintf(stderr, "utfstowcs(): mbstowcs() error\n");
	    ret = -1;
	}
    }

    return(ret);
}

/*  Compare numbers  */
int compnum(const void *n1, const void *n2)
{
    struct phnum *num1, *num2;

    num1 = (struct phnum *) n1;
    num2 = (struct phnum *) n2;

    return(strcmp(num1->num, num2->num));
}

/*  Read numbers from file  */
int readnums(FILE *fin, int mode)
{
    int ctrflg, ctrlen, fc, i, j, k, len, lenfld, numlen;
    char line[4096];
    struct phnum phonenum;

    ctrflg = 0;		/*  Country flag  */
    ctrlen = 0;		/*  Country code length  */
    fc = 0;		/*  Field count  */
    i = 0;
    j = 0;
    k = 0;
    len = 0;
    lenfld = 0;		/*  Length in length field  */
    numlen = 0;		/*  Number length  */
    memset(line, 0, 4096);
    memset(&phonenum, 0, sizeof(phonenum));

    /*
    *  Modes
    *  0  Phonebook
    *  1  Areas list
    *  2  Countries list
    *
    *  Formats
    *  Number <Tab> Name <Line Feed>
    *  Or
    *  Number <Tab> Name <Tab> Number length <Line Feed>
    */

    if (mode == 1 && country.name[0] != 0 && country.num[0] != 0) {
	/*  Areas list and country is set  */
	ctrlen = strlen(country.num);
    }

    while (fgets(line, 4096, fin)) {
	len = strlen(line);
	fc = 0;
	i = 0;
	j = 0;
	k = 0;
	lenfld = 0;
	numlen = 0;
	memset(&phonenum, 0, sizeof(phonenum));
	while (i < len) {
	    if (line[i] == 9) {
		/*  Tab  */
		line[i] = 0;	/*  Terminate  */
		if (fc == 0) {
		    /*  1st field; Number  */
		    if (mode == 0) {
			/*  Phone book; Don't alter  */
			strncpy(phonenum.num, line, ANL_NUM_MAX);
		    } else if (ctrlen != 0 && strncmp(line, country.num, \
		      ctrlen) == 0) {
			/*  Our country; Skip country code  */
			ctrflg = 1;
		 	if (country.trkpfx[0] == 0) {
			    /*  No area code prefix  */
			    strncpy(phonenum.num, line + ctrlen, ANL_NUM_MAX);
			} else {
			    /*  Insert area code prefix  */
			    snprintf(phonenum.num, ANL_NUM_MAX, "%s%s", \
			      country.trkpfx, line + ctrlen);
			}
		    } else {
			/*  Not our country  */
			ctrflg = 0;
			if (country.name[0] == 0) {
			    /*  No country selected; Assume 00  */
			    snprintf(phonenum.num, ANL_NUM_MAX, "00%s", \
			      line);
			} else if (country.incpfx[0] != 0) {
			    /*  Insert international call prexix  */
			    snprintf(phonenum.num, ANL_NUM_MAX, "%s%s", \
			      country.incpfx, line);
			} else {
			    /*  No prexix  */
			    strncpy(phonenum.num, line, ANL_NUM_MAX);
			}
		    }
		    phonenum.num[ANL_NUM_MAX] = 0;
		    j = i + 1;	/*  Points to name  */
		    numlen = strlen(phonenum.num);
		    if (mode == 0) {
			/*  Phonebook  */
			phonenum.len = 0;
		    } else {
			/*  Not a phonebook  */
			phonenum.len = numlen;
		    }
		} else if (mode == 0) {
		    /*  Extra Tab in phonebook  */
		    k = i + 1;	/*  Points to length  */
		}
		fc++;
		if (fc > 2) {
		    fclose(fin);
		    quit(1);
		    fprintf(stderr, "readnums(): Too many fields\n");
		    exit(1);
		}
	    } else if (line[i] < ' ' && line[i] > -1) {
		/*  Non-Tab Ctrl; Last field  */
		line[i] = 0;	/*  Terminate  */
		if (ctrflg == 0) {
		    /*  Not our country  */
		    strncpy(phonenum.name, line + j, ANL_NAME_MAX);
		} else {
		    /*  Our country; Remove ISO CC  */
		    strncpy(phonenum.name, line + j + 3, ANL_NAME_MAX);
		}
		phonenum.name[ANL_NAME_MAX] = 0;
		/*  Check for valid length field  */
		if (k != 0 && line[k] != 0 && (lenfld = atoi(line + k)) != 0) {
		    /*  Number range in phonebook  */
		    if (numlen == lenfld) {
			phonenum.len = numlen;
			if (debflg != 0) {
			    fprintf(stderr, \
			      "Debug: readnums(): Nr: %s Len: %d\n", \
			      line, numlen);
			}
		    } else {
			fclose(fin);
			quit(1);
			fprintf(stderr, \
			  "readnums(): Number length mismatch: %s %s\n", \
			  line, line + k);
			exit(1);
		    }
		}
		break;
	    }
	    i++;
	}	/*  End of while i < len  */
	memcpy(&(nums[numcnt]), &phonenum, sizeof(phonenum));
	numcnt++;
	if (numcnt > numax) {
	    if(debflg != 0)
		fprintf(stderr, "Debug: readnums(): growing nums\n");
	    initnums();
	}
    }	/*  End of while fgets line  */

    return(0);
}

/*  Write to phonebook file  */
int wrtphfil(int mode)
{
    int i, ptr;
    char *idxnam, *outnam;
    struct phidx idx;
    struct group *group;
    FILE *idxfil, *outfil;

    i = 0;
    ptr = 0;
    group  = NULL;
    idxfil = NULL;
    idxnam = NULL;
    outfil = NULL;
    outnam = NULL;
    memset(&idx, 0, sizeof(idx));

    /*  File names  */
    if (mode == 0) {
	/*  Contacts  */
	if (cmpconf != 1) {
	    /*  0 or 2 is no compact file format  */
	    if (debflg == 0) {
		outnam = ANL_NUMBERS_FILE;
	    } else {
		/*  Debug: Use default dir  */
		outnam = ANL_NUMS_FIL;
	    }
	} else {
	    /*  Compact file format  */
	    if (debflg == 0) {
		idxnam = ANL_NUMBERS_INDEX;
		outnam = ANL_NUMBERS_LIST;
	    } else {
		idxnam = ANL_NUMS_IDX;
		outnam = ANL_NUMS_LST;
	    }
	}
    } else if (mode == 1) {
	/*  Areas  */
	if (txtflg == 0) {
	    if (cmpconf != 1) {
		if (debflg == 0)
		    outnam = ANL_AREAS_FILE;
		else
		    outnam = ANL_AREA_FIL;
	    } else {
		if (debflg == 0) {
		    idxnam = ANL_AREAS_INDEX;
		    outnam = ANL_AREAS_LIST;
		} else {
		    idxnam = ANL_AREA_IDX;
		    outnam = ANL_AREA_LST;
		}
	    }
	} else {
	    /*  Text mode; Tsv out  */
	    outnam = ANL_AREA_TSV;
	}
    } else {
	/*  Mode is 2; Regions file  */
	if (txtflg == 0) {
	    if (cmpconf != 1) {
		if (debflg == 0)
		    outnam = ANL_REGIONS_FILE;
		else
		    outnam = ANL_REGS_FIL;
	    } else {
		if (debflg == 0) {
		    idxnam = ANL_REGIONS_INDEX;
		    outnam = ANL_REGIONS_LIST;
		} else {
		    idxnam = ANL_REGS_IDX;
		    outnam = ANL_REGS_LST;
		} 
	    }
	} else {
	    /*  Tsv out  */
	    outnam = ANL_REGS_TSV;
	}
    }

    /*  Open files  */
    if ((outfil = fopen(outnam, "w")) == NULL) {
	quit(1);
	fprintf(stderr, "wrtphfil(): Can't open %s\n", outnam);
	exit(1);
    }
    if (idxnam != NULL) {
	if ((idxfil = fopen(idxnam, "w")) == NULL) {
	    fclose(outfil);
	    quit(1);
	    fprintf(stderr, "wrtphfil(): Can't open %s\n", idxnam);
	    exit(1);
	}
    }

    i = 0;
    ptr = 0;
    /*  Nums is a global table of phone number structs  */
    while (i < numcnt) {
	if (mode == 0 || txtflg == 0) {
	    if (cmpconf != 1) {
		/*  Default file format  */
		fwrite(&(nums[i]), sizeof(struct phnum), 1, outfil);
	    } else {
		/*  Compact file format  */
		memset(&idx, 0, sizeof(idx));
		memcpy(idx.num, nums[i].num, ANL_NUM_SIZ);
		idx.len = nums[i].len;
		idx.offs = ptr;
		idx.namlen = strlen(nums[i].name);
		fwrite(&idx, sizeof(struct phidx), 1, idxfil);
		fwrite(&(nums[i].name), 1, idx.namlen + 1, outfil);
		ptr = ptr + idx.namlen + 1;
	    }
	} else {
	    /*  Text mode  */
	    fprintf(outfil, "%s\t%s\n", nums[i].num, nums[i].name);
	}
	i++;
    }

    if (idxfil != NULL) {
	fflush(idxfil);
	fclose(idxfil);
    }
    fflush(outfil);
    fclose(outfil);

    if (mode == 0 && geteuid() == 0) {
	/*  Phonebook and root; Set permissions  */
	if (grpnam[0] == 0) {
	    /*  Default group name  */
	    group = getgrnam("asterisk");
	} else {
	    /*  Group name set on command line  */
	    group = getgrnam(grpnam);
	}
	if (group != NULL) {
	    /* int chmod(const char *path, mode_t mode); */
	    chmod(outnam, S_IRUSR | S_IWUSR | S_IRGRP); /*  0640  */
	    /* int chown(const char *path, uid_t owner, gid_t group); */
	    chown(outnam, 0, group->gr_gid);            /*  root, asterisk  */
	    if (idxnam != NULL) {
		chmod(idxnam, S_IRUSR | S_IWUSR | S_IRGRP);
		chown(idxnam, 0, group->gr_gid);
	    }
	}
    }

    if (debflg != 0) {
	if (idxnam != NULL) {
	    fprintf(stderr, "Debug: wrtphfil(): Wrote %d entries to %s\n", \
	      i, idxnam);
	}
	fprintf(stderr, "Debug: wrtphfil(): Wrote %d entries to %s\n", \
	  i, outnam);
    }

    return(0);
}

/*  Write regions to file  */
int wrtregs(int mode)
{
    int i;
    char inam[256];
    FILE *fin;

    i = 0;
    numcnt = 0;		/*  Reset count  */
    fin = NULL;
    memset(inam, 0, 256);

    if (mode == 0) {
	if (debflg == 0)
	    sprintf(inam, "%s/region1.tsv", ANL_SRC_PATH);
	else
	    strcpy(inam, "region1.tsv");
    } else {
	if (debflg == 0)
	    sprintf(inam, "%s/region1fs.tsv", ANL_SRC_PATH);
	else
	    strcpy(inam, "region1fs.tsv");
    }

    if ((fin = fopen(inam, "r")) == NULL) {
	quit(1);
	fprintf(stderr, "wrtregs(): Can't open %s\n", inam);
	exit(1);
    }

    /*  Region 1  */
    readnums(fin, 2);
    fclose(fin);

    /*  Regions 2 - 9  */
    i = 2;
    while (i < 10) {
	if (debflg == 0)
	    sprintf(inam, "%s/region%d.tsv", ANL_SRC_PATH, i);
	else
	    sprintf(inam, "region%d.tsv", i);
	if ((fin = fopen(inam, "r")) == NULL) {
	    quit(1);
	    fprintf(stderr, "wrtregs(): Can't open %s\n", inam);
	    exit(1);
	}
	readnums(fin, 2);
	fclose(fin);
	i++;
    }

    /*  Sort numbers  */
    qsort(nums, numcnt, sizeof(struct phnum), compnum);

    /*  Write  */
    /* int wrtphfil(int mode) */
    wrtphfil(2);

    return(numcnt);
}

/*  Select regions  */
int selreg(void)
{
    int c, cont, mode, msdclk, msyptr, reg, ret;

    enum { NONE, NO_FS, WITH_FS, WRITE, QUIT };

    c = 0;
    cont = 1;
    msdclk = 0;
    msyptr = 0;
    ret = 0;

    if (regconf == 0) {
	/*  First time  */
	reg = 0;
    } else {
	/*  Not the first time  */
	reg = regconf - 1;
    }
    if(reg == 0)
	mode = 1;
    else
	mode = 2;

    drwmen(regmen, 1, 1, mode);
    if (reg == 0) {
	move(3, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
	move(4, 40);
	printw("[ ]");
    } else {
	move(3, 40);
	printw("[ ]");
	move(4, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
    }
    refresh();	/*  refresh() just before loop  */

    /*  Select region menu input  */
    while (c != EOF && cont != 0) {
	c = getch();
	switch (c) {
	    case KEY_DOWN:
		mode++;
		if(mode > QUIT)
		    mode = QUIT;
		break;
	    case KEY_UP:
		mode--;
		if(mode < NO_FS)
		    mode = NO_FS;
		break;
	    case KEY_PPAGE:
	    case KEY_HOME:
	    case '1':
		mode = NO_FS;
		break;
	    case '2':
		mode = WITH_FS;
		break;
	    case ' ':
		/*  Space toggles  */
		if (mode == NO_FS) {
		    mode = WITH_FS;
		    reg = 1;
		    clrline(6, 40, cols - 1);
		} else if (mode == WITH_FS) {
		    mode = NO_FS;
		    reg = 0;
		    clrline(6, 40, cols - 1);
		} else {
		    beep();
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: selreg(): Invalid input: %04X\n", c);
		    }
		}
		break;
	    case 'W':
	    case 'w':
		mode = WRITE;
		break;
	    case KEY_NPAGE:
	    case KEY_END:
	    case 'Q':
	    case 'q':
		mode = QUIT;
		break;
	    case 10:
	    case 13:
	    case KEY_ENTER:
		/*  Execute  */
	        if (mode == NO_FS) {
		    reg = 0;
		    /*  Erase file status  */
		    clrline(6, 40, cols - 1);
	        } else if (mode == WITH_FS) {
		    reg = 1;
		    /*  Erase file status  */
		    clrline(6, 40, cols - 1);
	        } else if (mode == WRITE) {
		    ret = wrtregs(reg);
		    regconf = reg + 1;	/*  Regconf set on write  */
		} else if (mode == QUIT) {
		    cont = 0;
	        }
		break;
	    case 27:
		mode = QUIT;
		cont = 0;
		break;
	    case KEY_MOUSE:
		msdclk = 0;
		msyptr = procmouse();
		msdclk = (msevent.bstate & BUTTON1_DOUBLE_CLICKED);
		if (msyptr == 3) {
		    mode = NO_FS;
		    if (msdclk != 0) {
			reg = 0;
			clrline(6, 40, cols - 1);
		    }
		} else if (msyptr == 4) {
		    mode = WITH_FS;
		    if (msdclk != 0) {
			reg = 1;
			clrline(6, 40, cols - 1);
		    }
		} else if (msyptr == 6) {
		    mode = WRITE;
		    if (msdclk != 0) {
			ret = wrtregs(reg);
			regconf = reg + 1;
		    }
		} else if (msyptr == 7) {
		    mode = QUIT;
		    if(msdclk != 0)
			cont = 0;
		}
		break;
	    case KEY_RESIZE:
		procresiz();
		break;
	    default:
		beep();
		if (debflg != 0) {
		    fprintf(stderr, \
		      "Debug: selreg(): Invalid input: %04X\n", c);
		}
		break;
	}	/*  End of switch  */

	drwmen(regmen, 1, 1, mode);
	if (reg == 0) {
	    move(3, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	    move(4, 40);
	    printw("[ ]");
	} else {
	    move(3, 40);
	    printw("[ ]");
	    move(4, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	}
	if (mode != WRITE) {
	    ret = 0;
	} else if (ret != 0) {
	    /*  Just wrote stuff  */
	    move(6, 40);
	    if (txtflg == 0) {
		if(cmpconf != 1)
		    printw("Wrote %d entries to regions.db", ret);
		else
		    printw("Wrote %d entries to regions.idx", ret);
	    } else {
		printw("Wrote %d entries to regions.tsv", ret);
	    }
	}
	refresh();	/*  refresh() at end of loop  */
    }

    numcnt = 0;

    return(0);
}

/*  Convert phonebook to binary  */
int convnums(void)
{
    int c;
    char inam[256];
    FILE *fin;

    c = 0;
    numcnt = 0;         /*  Reset count  */
    fin = NULL;
    memset(inam, 0, 256);

    /*  Prepare screen  */
    erase();
    box(stdscr, 0, 0);
    attrset(A_NORMAL);
    move(1, 1);

    /*  Open files  */
    fin = fopen("/etc/namelookup-agi/numbers.tsv", "r");
    if (fin == NULL) {
	/*  Try current dir  */
	fin = fopen("numbers.tsv", "r");
    }
    if (fin == NULL) {
	/*  Try homedir  */
	snprintf(inam, 255, "%s/numbers.tsv", home);
	fin = fopen(inam, "r");
    }

    if (fin == NULL) {
	printw("Can't open numbers.tsv");
	move(2, 1);
	printw("Put numbers.tsv in /etc/namelookup-agi/");
	move(3, 1);
	printw("Or start this program from the directory where ");
	printw("numbers.tsv is.");
	move(4, 1);
	printw("Or put numbers.tsv in your home directory.");

	move(6, 1);
	printw("Press any key to continue");
	refresh();
	c = getch();
	if (c == KEY_RESIZE)
	    procresiz();
	return(-1);
    }

    /*  Read numbers from file  */
    readnums(fin, 0);

    /*  Sort numbers  */
    qsort(nums, numcnt, sizeof(struct phnum), compnum);

    /*  Write; There is no text mode here  */
    wrtphfil(0);

    if (cmpconf != 1)
	printw("Wrote %d entries to numbers.db", numcnt);
    else
	printw("Wrote %d entries to numbers.idx", numcnt);
    numconf = 1;	/*  Numconf set on write  */
    move(3, 1);
    printw("Press any key to continue");
    refresh();
    c = getch();
    if (c == KEY_RESIZE)
	procresiz();

    numcnt = 0;

    return(0);
}

/*  Names compare  */
static int compnam(const void *n1, const void *n2)
{
    char *name1, *name2;

    name1 = (char *) n1;
    name2 = (char *) n2;

    return(strcmp(name1, name2));
}

/*  Convert blocklist to binary  */
int convblockl(void)
{
    int c, i, wlen;
    int blkcnt, blkentr, blkmax, blksiz, size;
    char inam[256], line[4096], name[ANL_NAME_SIZ], *outnam;
    wchar_t wname[ANL_NAME_SIZ];
    FILE *fin, *fout;
    struct blknam *blknams;

    c = 0;
    i = 0;
    blkcnt = 0;		/*  Number of names  */
    blkentr = 0;	/*  Number of names available  */
    blkmax = 0;		/*  Max number of names  */
    blksiz = 0;		/*  Size in bytes  */
    size = 0;
    wlen = 0;
    blknams = NULL;	/*  Blocklist table  */
    fin = NULL;
    fout = NULL;
    outnam = NULL;
    memset(inam, 0, 256);
    memset(line, 0, 4096);
    memset(name, 0, ANL_NAME_SIZ);
    memset(wname, 0, sizeof(wname));

    if (debflg == 0) {
	outnam = ANL_BLOCK_FILE;
    } else {
	outnam = ANL_BLCK_FIL;
    }

    /*  Prepare screen  */
    erase();
    box(stdscr, 0, 0);
    attrset(A_NORMAL);
    move(1, 1);

    /*  Allocate  */
    blkentr = ANL_DFLT_SIZ;	/*  Nr of entries available  */
    blkmax  = blkentr - 2;	/*  Maximum  */
    size = sizeof(struct blknam) * blkentr;
    blksiz = size;
    blknams = malloc((size_t) size);
    if (blknams == NULL) {
	quit(1);
	fprintf(stderr, \
	  "convblockl(): Can't allocate memory for names blocklist\n");
	exit(1);
    }
    memset(blknams, 0, size);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: convblockl(): Allocated %d bytes for %d names at %lx\n", \
	  size, blkentr, (long unsigned int) blknams);
	fprintf(stderr, "Debug: convblockl(): Max: %d\n", blkmax);
    }

    /*  Open files  */
    fin = fopen("/etc/namelookup-agi/agi-nameblock.txt", "r");
    if (fin == NULL) {
	/*  Try current dir  */
	fin = fopen("agi-nameblock.txt", "r");
    }
    if (fin == NULL) {
	/*  Try homedir  */
	snprintf(inam, 255, "%s/agi-nameblock.txt", home);
	fin = fopen(inam, "r");
    }
    if (fin != NULL)
	fout = fopen(outnam, "w");

    if (fin == NULL) {
	printw("Can't open agi-nameblock.txt");
	move(2, 1);
	printw("Put agi-nameblock.txt in /etc/namelookup-agi/");
	move(3, 1);
	printw("Or start this program from the directory where ");
	printw("agi-nameblock.txt is.");
	move(4, 1);
	printw("Or put agi-nameblock.txt in your home directory.");
    } else if (fout == NULL) {
	fclose(fin);
	fin = NULL;
	printw("Can't open %s", outnam);
    }

    if (fin == NULL) {
	move(6, 1);
	printw("Press any key to continue");
	refresh();
	c = getch();
	if (c == KEY_RESIZE)
	    procresiz();
	free(blknams);
	return(-1);
    }

    /*  Read names from file  */
    while (fgets(line, 4096, fin)) {
	/*  Cleanup  */
	i = 0;
	while (i < ANL_NAME_SIZ) {
	    if (utflg == 0 && line[i] < 0) {
		printw("You can only use non-ASCII in an UTF-8 environment.");
		move(2, 1);
		printw("Conversion aborted.");
		move(3, 1);
		printw("Press any key to continue");
		refresh();
		c = getch();
		if (c == KEY_RESIZE)
		    procresiz();
		free(blknams);
		return(-1);
	    }
	    if (line[i] < ' ' && line[i] > -1) {
		line[i] = 0;
		break;
	    }
	    i++;
	}
	if (line[0] == 0) {
	    /*  Empty line  */
	    continue;
	}
	line[ANL_NAME_MAX] = 0;

	/*  Convert to lower case  */
	if (utflg == 0) {
	    i = 0;
	    while (i < ANL_NAME_SIZ) {
		if(line[i] == 0)
		    break;
		line[i] = tolower(line[i]);
		i++;
	    }
	} else {
	    /*  Convert to wide characters  */
	    if ((wlen = mbstowcs(wname, line, ANL_NAME_MAX)) < 0) {
		free(blknams);
		quit(1);
		fprintf(stderr, "convblockl(): mbstowcs() failed\n");
        	exit(1);
            }
	    /*  Wide tolower  */
	    for(i = 0; i < wlen; i++)
		wname[i] = towlower(wname[i]);
	    /*  Convert to UTF-8  */
	    if (wcstombs(line, wname, ANL_NAME_MAX) < 0) {
		free(blknams);
		quit(1);
		fprintf(stderr, "convblockl(): wcstombs() failed\n");
		exit(1);
	    }
	}

	/*  Store  */
	strncpy(blknams[blkcnt].name, line, ANL_NAME_MAX);
	blkcnt++;
	if (blkcnt > blkmax) {
	    /*  Allocate more  */
	    size = 2 * blksiz;
	    blknams = realloc(blknams, (size_t) size);
	    if (blknams == NULL) {
		quit(1);
		fprintf(stderr, \
		  "convblockl(): Can't allocate memory for names blocklist\n");
		exit(1);
	    }
	    memset(blknams + blkentr, 0, blksiz);
	    /*  Set new values  */
	    blkentr = blkentr * 2;
	    blkmax  = blkentr - 2;
	    blksiz  = size;
	    if (debflg != 0) {
		fprintf(stderr, \
		  "Debug: convblockl(): Allocated %d bytes for %d names ", \
		  size, blkentr);
		fprintf(stderr, "at %lx\n", (long unsigned int) blknams);
		fprintf(stderr, "Debug: convblockl(): Max: %d\n", blkmax);
	    }
	}
    }
    fclose(fin);
    fin = NULL;

    /*  Sort names  */
    qsort(blknams, blkcnt, ANL_NAME_SIZ, compnam);

    /*  Write; There is no text mode here  */
    i = 0;
    while (i < blkcnt) {
        fwrite(&(blknams[i].name), ANL_NAME_SIZ, 1, fout);
        i++;
    }
    fflush(fout);
    fclose(fout);

    printw("Wrote %d entries to agi-nameblock.db", blkcnt);
    blkconf = 1;        /*  Numconf set on write  */
    move(3, 1);
    printw("Press any key to continue");
    refresh();
    c = getch();
    if (c == KEY_RESIZE)
        procresiz();

    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: convblockl(): Wrote %d entries to agi-nameblock.db\n", blkcnt);
    }

    free(blknams);
    return(0);
}

/*  Read countries file  */
int readcoutries(struct ctry *countries)
{
    int fc, i, j, len;
    char ch, inam[256], line[4096];
    FILE *fin;

    ch = 0;
    ctrycnt = 0;	/*  Number of countries  */
    fc = 0;
    i = 0;
    j = 0;
    len = 0;
    fin = NULL;
    memset(inam, 0, 256);
    memset(line, 0, 4096);

    if(debflg == 0)
	snprintf(inam, 255, "%s/countries.tsv", ANL_SRC_PATH);
    else
	strcpy(inam, "countries.tsv");

    if ((fin = fopen(inam, "r")) == NULL) {
	quit(0);	/*  Not yet in ncurses  */
	fprintf(stderr, "readcoutries(): Cant open %s\n", inam);
	exit(1);
    }

    while (fgets(line, 4096, fin)) {
	if (line[0] == '#') {
	    /*  Remark  */
	    continue;
	}
	ch = 0;
	fc = 0;
	i = 0;
	j = 0;
	len = strlen(line);
	while (i < len) {
	    if (line[i] == 9 || (line[i] == ' ' && line[i + 1] == '#') || \
	      line[i] == '#' || (line[i] < ' ' && line[i] > 0)) {
		/*  Tab, Remark, Ctrl or Lf  */
		ch = line[i];
		line[i] = 0;
		if (fc == 0) {
		    /*  Country  */
		    if (utfstowcs(line, j - 1, countries[ctrycnt].name, \
		      ANL_CTRY_NAM_MAX) < 0) {
			quit(0);
			fclose(fin);
			fprintf(stderr, "readcoutries(): utfstowcs error\n");
			exit(1);
		    }
		} else if (fc == 1) {
		    /*  ISO CC, used to be country code  */
		    strncpy(countries[ctrycnt].cc, line + j, ANL_CTRY_CC_MAX);
		    if (countries[ctrycnt].cc[0] < 'A') {
			quit(0);
			fclose(fin);
			fprintf(stderr, \
			  "readcoutries(): countries.tsv file error:\n");
			fprintf(stderr, \
			  "The second field should be an ISO CC.\n");
			exit(1);
		    }
		} else if (fc == 2) {
		    /*  Country code  */
		    strncpy(countries[ctrycnt].num, line + j, ANL_CTRY_NUM_MAX);
		} else if (fc == 3) {
		    /*  Int prefix  */
		    strncpy(countries[ctrycnt].incpfx, line + j, \
		      ANL_CTRY_INC_MAX);
		} else if (fc == 4) {
		    /*  Trunk prefix  */
		    strncpy(countries[ctrycnt].trkpfx, line + j, \
		      ANL_CTRY_TRK_MAX);
		}
		if (ch != 9) {
		    /*  Not a Tab  */
		    break;
		}
		j = i + 1;
		fc++;
	    }
	    i++;
	}
	ctrycnt++;
	if (ctrycnt > ANL_COUNTRIES_MAX) {
	    quit(0);
	    fclose(fin);
	    fprintf(stderr, "readcoutries(): Too many countries\n");
	    exit(1);
	}
    }

    fclose(fin);

    if (debflg != 0) {
	fprintf(stderr, "Debug: readcountries(): read %d countries\n", \
	  ctrycnt);
    }

    return(ctrycnt);
}

/*  Select country  */
int selctry(char *isostr)
{
    int c, cnt, curr, i, max, mode, msdclk, msyptr, offs;

    /*  None is actually a mode here  */
    enum { SCROLL, NONE, QUIT };

    c = 0;
    cnt = 0;    /*  Number of countries  */
    curr = 0;   /*  Selected coutry in list  */
    i = 0;
    max = 0;    /*  Rows - 10  */
    mode = 0;	/*  Defaults to scroll  */
    msdclk = 0;
    msyptr = 0;
    offs = 0;   /*  Varies between 0 and max  */

    /*  Count is one bigger then maximum value of current  */
    cnt = ctrycnt - 1;
    if (cnt < 0) {
	/*  This shouldn't happen  */
	cnt = 0;
    }

    if (country.name[0] != 0 && country.num[0] != 0) {
	/*  Go to selected  */
	i = 0;
	while (i <= cnt) {
	    if (strcmp(country.num, countries[i].num) == 0 && \
	      strcmp(country.cc, countries[i].cc) == 0) {
		curr = i;
		break;
	    }
	    i++;
	}
    } else {
	/*  Use locale ISO CC  */
	i = 0;
	while (i <= cnt) {
	    if (strcasecmp(isostr, countries[i].cc) == 0) {
		curr = i;
		break;
	    }
	    i++;
	}
    }

    /*  Calculate offset  */
    max = rows - 10;
    if(max < 0)
	max = 0;
    offs = curr - max / 2;
    if(offs > cnt - max)
	offs = cnt - max;
    if(offs < 0)
	offs = 0;
    if (debflg != 0) {
	fprintf(stderr, "Debug: selctry(): Curr %d Offs %d Max %d\n", \
	  curr, offs, max);
    }

    /*  Menu at bottom  */
    drwmen(cntrymen, rows - 3, 1, 0);
    /*  Lines  */
    mvhline(2, 1, 0, cols - 2);
    mvaddch(2, 0, ACS_LTEE);
    mvaddch(2, cols - 1, ACS_RTEE);
    mvhline(rows - 4, 1, 0, cols - 2);
    mvaddch(rows - 4, 0, ACS_LTEE);
    mvaddch(rows - 4, cols - 1, ACS_RTEE);
    /*  Country list  */
    /* void drwctry(struct ctry *countries, int offs, int curr, int max, int mode) */
    drwctry(countries, offs, curr, max, mode);
    /* void drwbar(int y, int x, int curr, int cnt, int max) */
    drwbar(5, cols - 2, curr, cnt, max);
    refresh();

    /*  Read user input  */
    while (c != EOF) {
	c = getch();
	if (c > '`' && c < '{') {
	    /*  To upper  */
	    c -= 32;
	}
	if (c == KEY_DOWN) {
	    if (mode == SCROLL) {
		curr++;
		if(curr > cnt)
		    curr = cnt;
	    } else {
		mode++;
		if(mode > QUIT)
		    mode = QUIT;
	    }
	} else if (c == KEY_UP) {
	    if (mode == SCROLL) {
		curr--;
		if(curr < 0)
		    curr = 0;
	    } else {
		mode--;
		if(mode < NONE)
		    mode = NONE;
	    }
	} else if (c == KEY_NPAGE) {
	    if (mode == SCROLL) {
		curr += max;
		if(curr > cnt)
		    curr = cnt;
	    } else {
		mode = QUIT;
	    }
	} else if (c == KEY_PPAGE) {
	    if (mode == SCROLL) {
		curr -= max;
		if(curr < 0)
		    curr = 0;
	    } else {
		mode = NONE;
	    }
	} else if (c == KEY_END) {
	    if (mode == SCROLL) {
		curr = cnt;
	    } else {
		mode = QUIT;
	    }
	} else if (c == KEY_HOME) {
	    if (mode == SCROLL) {
		curr = 0;
	    } else {
		mode = NONE;
	    }
	} else if (c == 9 || c == KEY_STAB) {
	    /*  Toggle select function  */
	    if(mode == 0)
		mode = QUIT;
	    else
		mode = SCROLL;
	} else if (c == 'N' || c == 'n') {
	    mode = NONE;
	} else if (c == 'Q' || c == 'q') {
	    mode = QUIT;
	} else if (c > '@' && c < '[') {
	    /*  Look for country starting with  */
	    i = curr;
	    if (countries[curr].name[0] < c) {
		while (i <= cnt) {
		    if(countries[i].name[0] == c)
			break;
		    i++;
		}
	    } else if (countries[curr].name[0] > c) {
		while (i >= 0) {
		    if(countries[i].name[0] == c)
			break;
		    i--;
		}
	    }
	    if (i >= 0 && i <= cnt) {
		/*  Found  */
		curr = i;
		mode = SCROLL;
	    } else {
		beep();
		if (debflg != 0) {
		    fprintf(stderr, \
		      "Debug: selctry(): Invalid input: %04X\n", c);
		}
	    }
	} else if (c == 10 || c == 13 || c == KEY_ENTER) {
	    if (mode == 0) {
		memcpy(&country, &(countries[curr]), sizeof(struct ctry));
	    } else if (mode == NONE) {
		/*  None  */
		memset(&country, 0, sizeof(country));
		mode = SCROLL;
	    } else if (mode == QUIT) {
		break;
	    }
	} else if (c == ' ' && mode == 0) {
	    /*  Space toggles  */
	    if(country.num[0] == 0)
		memcpy(&country, &(countries[curr]), sizeof(struct ctry));
	    else
		memset(&country, 0, sizeof(country));
	} else if (c == 27) {
	    break;
	} else if (c == KEY_MOUSE) {
	    msdclk = 0;
	    msyptr = procmouse();
	    msdclk = (msevent.bstate & BUTTON1_DOUBLE_CLICKED);
	    if (msyptr == 4) {
		/*  Up  */
		curr -= max / 2;
		if(curr < 0)
		    curr = 0;
		mode = SCROLL;
	    } else if (msyptr == rows - 4) {
		/*  Down  */
		curr += max / 2;
		if (curr > cnt)
		    curr = cnt;
		mode = SCROLL;
	    } else if (msyptr > 4 && msyptr < rows - 4) {
		/*  This doesn't change offset  */
		curr = offs + msyptr - 5;
		mode = SCROLL;
		if(msdclk != 0)
		    memcpy(&country, &(countries[curr]), sizeof(struct ctry));
	    } else if (msyptr == rows - 3) {
		mode = NONE;
		if (msdclk != 0) {
		    memset(&country, 0, sizeof(country));
		    mode = SCROLL;
		}
	    } else if (msyptr == rows - 2) {
		mode = QUIT;
		if(msdclk != 0)
		    break;
	    }
	} else if (c == KEY_RESIZE) {
	    /*  Offs varies between 0 and max  */
	    procresiz();
	    max = rows - 10;
	    if(max < 0)
		max = 0;
	} else {
	    beep();
	    if (debflg != 0) {
		fprintf(stderr, \
		  "Debug: selctry(): Invalid input: %04X\n", c);
	    }
	}

	/*  Calculate offset  */
	max = rows - 10;
	if(max < 0)
	    max = 0;
	offs = curr - max / 2;
	if(offs > cnt - max)
	    offs = cnt - max;
	if(offs < 0)
	    offs = 0;

	/*  Bottom menu  */
	drwmen(cntrymen, rows - 3, 1, mode);
	/*  Lines  */
	mvhline(2, 1, 0, cols - 2);
	mvaddch(2, 0, ACS_LTEE);
	mvaddch(2, cols - 1, ACS_RTEE);
	mvhline(rows - 4, 1, 0, cols - 2);
	mvaddch(rows - 4, 0, ACS_LTEE);
	mvaddch(rows - 4, cols - 1, ACS_RTEE);
	/*  Country list  */
	drwctry(countries, offs, curr, max, mode);
	/*  Scroll bar  */
	drwbar(5, cols - 2, curr, cnt, max);
	refresh();
    }

    return(0);
}

/*  Write configuration to agi-namelookup.conf  */
int wrtconf(char *confnam, int weblkp, int cmpfil)
{
    FILE *confil;

    confil = NULL;

    if ((confil = fopen(confnam, "w")) == NULL) {
	quit(1);
	fprintf(stderr, "wrtconf(): Can't open %s\n", confnam);
	exit(1);
    }

    if(weblkp == 1)
	fprintf(confil, "weblookup=on\n");
    else if(weblkp == 2)
	fprintf(confil, "weblookup=off\n");

    if(cmpfil == 1)
	fprintf(confil, "compact=on\n");
    else if(cmpfil == 2)
	fprintf(confil, "compact=off\n");

    fclose(confil);

    if(debflg != 0)
	fprintf(stderr, "Debug: wrtconf(): Wrote agi-namelookup.conf\n");

    return(weblkp);
}

/*  General configuration  */
int genconf(void)
{
    int c, cmpfil, cont, mode, msdclk, msyptr, ret, weblkp;
    char *confnam;

    enum { NONE, WEB_LKP, NO_WEB, CMP_FIL, NO_CMP, WRITE, QUIT };

    c = 0;
    cont = 1;
    mode = 1;
    msdclk = 0;
    msyptr = 0;
    ret = 0;
    confnam = NULL;

    /*  Copy from global  */
    weblkp = webconf;
    cmpfil = cmpconf;

    if (debflg == 0)
	confnam = ANL_CONFIG_FILE;
    else
	confnam = ANL_CONF_FIL;

    if (weblkp == 0) {
	/*  First time  */
	weblkp = 1;     /*  Default is enabled  */
    } else if (weblkp == 2) {
	/*  Disabled in config file  */
	mode = 2;
    }
    if (cmpfil == 0) {
	/*  First time  */
	cmpfil = 2;	/*  Default is disabled  */
    }

    drwmen(confmen, 1, 1, weblkp);
    if (weblkp == 1) {
	move(3, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
	move(4, 40);
	printw("[ ]");
    } else if (weblkp == 2) {
	move(3, 40);
	printw("[ ]");
	move(4, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
    }
    if (cmpfil == 1) {
	move(5, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
	move(6, 40);
	printw("[ ]");
    } else if (cmpfil == 2) {
	move(5, 40);
	printw("[ ]");
	move(6, 40);
	if(colflg != 0)
	    attron(COLOR_PAIR(highlt));
	printw("[X]");
	if(colflg != 0)
	    attroff(COLOR_PAIR(highlt));
    }
    refresh();

    /*  General configuration menu input  */
    while (c != EOF && cont != 0) {
	c = getch();
	switch (c) {
	    case KEY_DOWN:
		mode++;
		if(mode > QUIT)
		    mode = QUIT;
		break;
	    case KEY_UP:
		mode--;
		if(mode < WEB_LKP)
		    mode = WEB_LKP;
		break;
	    case KEY_PPAGE:
	    case KEY_HOME:
	    case '1':
		mode = WEB_LKP;
		break;
	    case '2':
		mode = NO_WEB;
		break;
	    case '3':
		mode = CMP_FIL;
		break;
	    case '4':
		mode = NO_CMP;
		break;
	    case ' ':
		/*  Space toggles  */
		if (mode == WEB_LKP) {
		    mode = NO_WEB;
		    weblkp = 2;
		    clrline(8, 40, cols - 1);
		} else if(mode == NO_WEB) {
		    mode = WEB_LKP;
		    weblkp = 1;
		    clrline(8, 40, cols - 1);
		} else if (mode == CMP_FIL) {
		    mode = NO_CMP;
		    cmpfil = 2;
		    clrline(8, 40, cols - 1);
		} else if (mode == NO_CMP) {
		    mode = CMP_FIL;
		    cmpfil = 1;
		    clrline(8, 40, cols - 1);
		} else {
		    beep();
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: genconf(): Invalid input: %04X\n", c);
		    }
		}
		break;
	    case 'W':
	    case 'w':
		mode = WRITE;
		break;
	    case KEY_NPAGE:
	    case KEY_END:
	    case 'Q':
	    case 'q':
		mode = QUIT;
		break;
	    case 10:
	    case 13:
	    case KEY_ENTER:
		/*  Execute  */
		if (mode == WEB_LKP) {
		    weblkp = 1;
		    /*  Erase file status  */
		    clrline(8, 40, cols - 1);
		} else if (mode == NO_WEB) {
		    weblkp = 2;
		    clrline(8, 40, cols - 1);
		} else if (mode == CMP_FIL) {
		    cmpfil = 1;
		    clrline(8, 40, cols - 1);
		} else if (mode == NO_CMP) {
		    cmpfil = 2;
		    clrline(8, 40, cols - 1);
		} else if (mode == WRITE) {
		    /* int wrtconf(char *confnam, int weblkp, int cmpfil) */
		    ret = wrtconf(confnam, weblkp, cmpfil);
		    /*  Copy to global vars  */
		    webconf = weblkp;
		    cmpconf = cmpfil;
		} else if (mode == QUIT) {
		    cont = 0;
		}
		break;
	    case 27:
		mode = QUIT;
		cont = 0;
		break;
	    case KEY_MOUSE:
		msdclk = 0;
		msyptr = procmouse();
		msdclk = (msevent.bstate & BUTTON1_DOUBLE_CLICKED);
		if (msyptr == 3) {
		    mode = WEB_LKP;
		    if (msdclk != 0) {
			weblkp = 1;
			clrline(8, 40, cols - 1);
		    }
		} else if (msyptr == 4) {
		    mode = NO_WEB;
		    if (msdclk != 0) {
			weblkp = 2;
			clrline(8, 40, cols - 1);
		    }
		} else if (msyptr == 5) {
		    mode = CMP_FIL;
		    if (msdclk != 0) {
			cmpfil = 1;
			clrline(8, 40, cols - 1);
		    }
		} else if (msyptr == 6) {
		    mode = NO_CMP;
		    if (msdclk != 0) {
			cmpfil = 2;
			clrline(8, 40, cols - 1);
		    }
		} else if (msyptr == 8) {
		    mode = WRITE;
		    if (msdclk != 0) {
			/* int wrtconf(char *confnam, int weblkp, int cmpfil) */
			ret = wrtconf(confnam, weblkp, cmpfil);
			/*  Copy to global vars  */
			webconf = weblkp;
			cmpconf = cmpfil;
		    }
		} else if (msyptr == 9) {
		    mode = QUIT;
		    if(msdclk != 0)
			cont = 0;
		}
		break;
	    case KEY_RESIZE:
		procresiz();
		break;
	    default:
		beep();
		if (debflg != 0) {
		    fprintf(stderr, \
		      "Debug: genconf(): Invalid input: %04X\n", c);
		}
		break;
	}	/* end of switch  */

	drwmen(confmen, 1, 1, mode);
	if (weblkp == 1) {
	    move(3, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	    move(4, 40);
	    printw("[ ]");
	} else if (weblkp == 2) {
	    move(3, 40);
	    printw("[ ]");
	    move(4, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	}
	if (cmpfil == 1) {
	    move(5, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	    move(6, 40);
	    printw("[ ]");
	} else if (cmpfil == 2) {
	    move(5, 40);
	    printw("[ ]");
	    move(6, 40);
	    if(colflg != 0)
		attron(COLOR_PAIR(highlt));
	    printw("[X]");
	    if(colflg != 0)
		attroff(COLOR_PAIR(highlt));
	}

	if (mode != WRITE) {
	    ret = 0;
	} else  if (ret != 0) {
	    /*  Just wrote stuff  */
	    move(8, 40);
	    printw("Wrote agi-namelookup.conf");
	}
	refresh();
    }

    return(0);
}

/*  Write areas db or tsv  */
int wrtareas(struct area *areas, int cnt)
{
    int i;
    char inam[256];
    FILE *fin;

    i = 0;
    numcnt = 0;		/*  Reset count  */
    fin = NULL;
    memset(inam, 0, 256);

    /*  Read the selected areas  */
    i = 0;
    while (i <= cnt) {
	if (areas[i].sel != 0) {
	    if (debflg == 0)
		snprintf(inam, 255, "%s/%s", ANL_SRC_PATH, areas[i].filnam);
	    else
		strncpy(inam, areas[i].filnam, 255);
	    if ((fin = fopen(inam, "r")) == NULL) {
		quit(1);
		fprintf(stderr, "wrtareas(): Can't open %s\n", inam);
		exit(1);
	    }
	    readnums(fin, 1);
	    fclose(fin);
	}
	i++;
    }

    /*  Sort numbers  */
    qsort(nums, numcnt, sizeof(struct phnum), compnum);

    /*  Write  */
    wrtphfil(1);

    return(numcnt);
}

/*  Read area files list file  */
int readareas(struct area *areas)
{
    int fc, i, isocnt, j, len, ret;
    char ch, filnam[256], iso[4], line[4096], lng[4], res[4];
    FILE *fin;

    areacnt = 0;
    ch = 0;
    fc = 0;
    i = 0;
    isocnt = 0;
    j = 0;
    len = 0;
    ret = 0;
    fin = NULL;
    memset(filnam, 0, 256);
    memset(iso, 0, 4);
    memset(line, 0, 4096);
    memset(lng, 0, 4);
    memset(res, 0, 4);

    if(debflg == 0)
	snprintf(filnam, 255, "%s/areas-list.tsv", ANL_SRC_PATH);
    else
	strcpy(filnam, "areas-list.tsv");

    if ((fin = fopen(filnam, "r")) == NULL) {
	quit(0);	/*  Not yet in ncurses  */
	fprintf(stderr, "readareas(): Can't open %s\n", filnam);
	exit(1);
    }

    while (fgets(line, 4096, fin)) {
	if (line[0] == '#') {
	    /*  Remark  */
	    continue;
	}
	ch = 0;
	fc = 0;
	i = 0;
	j = 0;
	len = strlen(line);
	while (i < len) {
	    if (line[i] == 9 || (line[i] == ' ' && line[i + 1] == '#') || \
	      line[i] == '#' || (line[i] < ' ' && line[i] > 0)) {
		/*  Tab, Remark, Ctrl or Lf  */
		ch = line[i];
		line[i] = 0;
		if (fc == 0) {
		    /*  ISO CC length is always 2  */
		    strncpy(areas[areacnt].iso, line, 2);
		    if (strcmp(areas[areacnt].iso, iso) != 0) {
			/*  New ISO CC  */
			isocnt = 0;
		    }
		    strncpy(iso, line, 2);
		} else if (fc == 1) {
		    /*  File name  */
		    strncpy(areas[areacnt].filnam, line + j, ANL_AR_FNAM_MAX);
		} else if (fc == 2) {
		    /* int utfstowcs(char *inbuf, int inlen, \
		      wchar_t *outbuf, int outsiz) */
		    ret = utfstowcs(line + j, i - j, areas[areacnt].country, \
		      ANL_AR_CTRY_MAX);
		} else if (fc == 3) {
		    ret = utfstowcs(line + j, i - j, areas[areacnt].lang, \
		      ANL_AR_LANG_MAX);
		} else if (fc == 4) {
		    ret = utfstowcs(line + j, i - j, areas[areacnt].res, \
		      ANL_AR_RES_MAX);
		}
		if (ret < 0) {
		    quit(0);
		    fclose(fin);
		    fprintf(stderr, "readareas(): utfstowcs error\n");
		    exit(1);
		}
		if (ch != 9) {
		    /*  Not a Tab  */
		    break;
		}
		j = i + 1;
		fc++;
	    }	/*  End of field processing  */
	    i++;
	}	/*  End of while i < len  */
	areas[areacnt].isocnt = isocnt;
	if(strcmp(areas[areacnt].iso, iso) == 0)
	    isocnt++;
	else
	    isocnt = 0;
	strcpy(iso, areas[areacnt].iso);

	areacnt++;

    }	/*  End of while fgets()  */
    fclose(fin);

    if(debflg != 0)
	fprintf(stderr, "Debug: readareas(): Read %d entries\n", areacnt);

    return(areacnt);
}

/*  Select area files  */
int selarea(void)
{
    int c, cnt, curr, i, max, mode, msdclk, msyptr, offs, ret;

    enum { SCROLL, ALLAR, NOAR, WRITE, QUIT };

    c = 0;
    cnt = 0;
    curr = 0;
    i = 0;
    max = 0;
    mode = 0;	/*  Defaults to scroll  */
    msdclk = 0;
    msyptr = 0;
    offs = 0;
    ret = 0;

    max = rows - 12;
    if(max < 0)
	max = 0;
    /*  Count is one bigger then maximum value of current  */
    cnt = areacnt - 1;
    if (cnt < 0) {
	/*  This shouldn't happen  */
	cnt = 0;
    }

    /*  Menu at bottom  */
    drwmen(areamen, rows - 5, 1, 0);
    /*  Lines  */
    mvhline(2, 1, 0, cols - 2);
    mvaddch(2, 0, ACS_LTEE);
    mvaddch(2, cols - 1, ACS_RTEE);
    mvhline(rows - 6, 1, 0, cols - 2);
    mvaddch(rows - 6, 0, ACS_LTEE);
    mvaddch(rows - 6, cols - 1, ACS_RTEE);
    /*  Areas menu  */
    drwareas(areas, 0, 0, max, mode);
    /* void drwbar(int y, int x, int curr, int cnt, int max) */
    drwbar(5, cols - 2, curr, cnt, max);
    refresh();

    while (c != EOF) {
	c = getch();
	if (c > '`' && c < '{') {
	    /* To upper  */
	    c -= 32;
	}
	if (c == KEY_DOWN) {
	    if (mode == SCROLL) {
		curr++;
		if(curr > cnt)
		    curr = cnt;
	    } else {
		mode++;
		if(mode > QUIT)
		    mode = QUIT;
	    }
	} else if (c == KEY_UP) {
	    if (mode == SCROLL) {
		curr--;
		if(curr < 0)
		    curr = 0;
	    } else {
		mode--;
		if(mode < ALLAR)
		    mode = ALLAR;
	    }
	} else if (c == KEY_NPAGE) {
	    if (mode == SCROLL) {
		curr += max;
		if(curr > cnt)
		    curr = cnt;
	    } else {
		mode = QUIT;
	    }
	} else if (c == KEY_PPAGE) {
	    if (mode == SCROLL) {
		curr -= max;
		if(curr < 0)
		    curr = 0;
	    } else {
		mode = ALLAR;
	    }
	} else if (c == KEY_END) {
	    if (mode == SCROLL) {
		curr = cnt;
	    } else {
		mode = QUIT;
	    }
	} else if (c == KEY_HOME) {
	    if (mode == SCROLL) {
		curr = 0;
	    } else {
		mode = ALLAR;
	    }
	} else if (c == KEY_STAB || c == 9) {
	    /*  Toggle select function  */
	    if(mode == 0)
		mode = WRITE;
	    else
		mode = SCROLL;
	} else if (c == 'A' || c == 'a') {
	    mode = ALLAR;
	} else if (c == 'N' || c == 'n') {
	    mode = NOAR;
	} else if (c == 'W' || c == 'w') {
	    mode = WRITE;
	} else if (c == 'Q' || c == 'q') {
	    mode = QUIT;
	} else if (c > '@' && c < '[') {
	    /*  Look for country starting with  */
	    i = curr;
	    if (areas[curr].country[0] < c) {
		while (i <= cnt) {
		    if(areas[i].country[0] == c)
			break;
		    i++;
		}
	    } else if (areas[curr].country[0] > c) {
		while (i >= 0) {
		    if(areas[i].country[0] == c)
			break;
		    i--;
		}
	    }
	    //if (i > 0 && i < cnt) {
	    if (i >= 0 && i <= cnt) {
		/*  Found  */
		curr = i;
		mode = SCROLL;
	    } else {
		beep();
		if (debflg != 0) {
		    fprintf(stderr, \
		      "Debug: selarea(): Invalid input: %04X\n", c);
		}
	    }
	} else if (c == ' ' && mode == 0) {
	    /*  Space toggles  */
	    if (areas[curr].sel == 0) {
		i = curr - areas[curr].isocnt;
		areas[i].sel = 0;
		i++;
		while (areas[i].isocnt != 0) {
		    areas[i].sel = 0;
		    i++;
		    if(i > cnt)
			break;
		}
		areas[curr].sel = 1;
	    } else {
		areas[curr].sel = 0;
	    }
	} else if (c == 10 || c == 13 || c == KEY_ENTER) {
	    /*  Execute  */
	    if (mode == 0) {
		/*  Select or deselect  */
		if (areas[curr].sel == 0) {
		    /*  Clear this ISO  */
		    i = curr - areas[curr].isocnt;
		    areas[i].sel = 0;
		    i++;
		    while (areas[i].isocnt != 0) {
			areas[i].sel = 0;
			i++;
			if(i > cnt)
			    break;
		    }
		    areas[curr].sel = 1;
		} else {
		    /*  Deselect  */
		    areas[curr].sel = 0;
		}
	    } else if (mode == ALLAR) {
		i = 0;
		while (i <= cnt) {
		    if(areas[i].isocnt == 0)
			areas[i].sel = 1;
		    else
			areas[i].sel = 0;
		    i++;
		}
	    } else if(mode == NOAR) {
		i = 0;
		while (i <= cnt) {
		    areas[i].sel = 0;
		    i++;
		}
	    } else if (mode == WRITE) {
		ret = wrtareas(areas, cnt);
		arconf = 2;	/*  2: Arconf is set on write  */
	    } else if (mode == QUIT) {
		break;
	    }
	} else if (c == 27) {
	    break;
	} else if (c == KEY_MOUSE) {
	    msdclk = 0;
	    msyptr = procmouse();
	    msdclk = (msevent.bstate & BUTTON1_DOUBLE_CLICKED);
	    if (msyptr == 4) {
		/*  Up  */
		curr -= max / 2;
		if(curr < 0)
		    curr = 0;
		mode = SCROLL;
	    } else if (msyptr == rows - 6) {
		/*  Down  */
		curr += max / 2;
		if (curr > cnt)
		    curr = cnt;
		mode = SCROLL;
	    } else if (msyptr > 4 && msyptr < rows - 6) {
		/*  This doesn't change offset  */
		curr = offs + msyptr - 5;
		mode = SCROLL;
		if (msdclk != 0) {
		    /*  Select or deselect  */
		    if (areas[curr].sel == 0) {
			/*  Clear this ISO  */
			i = curr - areas[curr].isocnt;
			areas[i].sel = 0;
			i++;
			while (areas[i].isocnt != 0) {
			    areas[i].sel = 0;
			    i++;
			    if(i > cnt)
				break;
			}
			areas[curr].sel = 1;
		    } else {
			/*  Deselect  */
			areas[curr].sel = 0;
		    }
		}
	    } else if (msyptr == rows - 5) {
		mode = ALLAR;
		if (msdclk != 0) {
		    i = 0;
		    while (i <= cnt) {
			if(areas[i].isocnt == 0)
			    areas[i].sel = 1;
			else
			    areas[i].sel = 0;
			i++;
		    }
		}
	    } else if (msyptr == rows - 4) {
		mode = NOAR;
		if (msdclk != 0) {
		    i = 0;
		    while (i <= cnt) {
			areas[i].sel = 0;
			i++;
		    }
		}
	    } else if (msyptr == rows - 3) {
		mode = WRITE;
		if (msdclk != 0) {
		    ret = wrtareas(areas, cnt);
		    arconf = 2;	/*  2: Arconf is set on write  */
		}
	    } else if (msyptr == rows - 2) {
		mode = QUIT;
		if(msdclk != 0)
		    break;
	    }
	} else if (c == KEY_RESIZE) {
	    procresiz();
	    max = rows - 12;
	    if(max < 0)
		max = 0;
	} else {
	    beep();
	    if (debflg != 0) {
		fprintf(stderr, \
		  "Debug: selarea(): Invalid input: %04X\n", c);
	    }
	}

	/*  Calculate offset  */
	max = rows - 12;
	if(max < 0)
	    max = 0;
	offs = curr - max / 2;
	if(offs > cnt - max)
	    offs = cnt - max;
	if(offs < 0)
	    offs = 0;

	drwmen(areamen, rows - 5, 1, mode);
	mvhline(2, 1, 0, cols - 2);
	mvaddch(2, 0, ACS_LTEE);
	mvaddch(2, cols - 1, ACS_RTEE);
	mvhline(rows - 6, 1, 0, cols - 2);
	mvaddch(rows - 6, 0, ACS_LTEE);
	mvaddch(rows - 6, cols - 1, ACS_RTEE);
	drwareas(areas, offs, curr, max, mode);
	drwbar(5, cols - 2, curr, cnt, max);
	if (mode != WRITE) {
	    ret = 0;
	} else if (ret != 0) {
	    move(rows - 3, 40);
	    if (txtflg == 0) {
		if(cmpconf != 1)
		    printw("Wrote %d entries to areas.db", ret);
		else
		    printw("Wrote %d entries to areas.idx", ret);
	    } else {
		printw("Wrote %d entries to areas.tsv", ret);
	    }
	}
	refresh();
    }

    numcnt = 0;

    return(0);
}

/*  Read namelookup.agi config file  */
int readagiconf(void)
{
    char inam[256], line[4096];
    int i, j;
    FILE *infil;

    i = 0;
    j = 0;
    infil = NULL;
    memset(inam, 0, 256);
    memset(line, 0, 4096);

    /*  Check namelookup.agi conf file first  */
    if (debflg == 0)
	strcpy(inam, ANL_CONFIG_FILE);
    else
	strcpy(inam, ANL_CONF_FIL);
    if ((infil = fopen(inam, "r")) != NULL) {
	/*  File exists  */
	while (fgets(line, 4096, infil)) {
	    if (strncasecmp(line, "weblookup=", 10) == 0) {
		if (line[10] == '0' || \
		  strncasecmp(line + 10, "no", 2) == 0 || \
		  strncasecmp(line + 10, "off", 3) == 0 || \
		  strncasecmp(line + 10, "false", 5) == 0) {
		    /*  Weblookups are disabled  */
		    webconf = 2;
		} else if (line[10] == '1' || \
		  strncasecmp(line + 10, "yes", 3) == 0 || \
		  strncasecmp(line + 10, "on", 2) == 0 || \
		  strncasecmp(line + 10, "true", 4) == 0) {
		    /*  Weblookups are enabled  */
		    webconf = 1;
		}
	    } else if (strncasecmp(line, "compact=", 8) == 0) {
		if (line[8] == '0' || \
		  strncasecmp(line + 8, "no", 2) == 0 || \
		  strncasecmp(line + 8, "off", 3) == 0 || \
		  strncasecmp(line + 8, "false", 5) == 0) {
		    /*  Compact files are disabled  */
		    cmpconf = 2;
		} else if (line[8] == '1' || \
		  strncasecmp(line + 8, "yes", 3) == 0 || \
		  strncasecmp(line + 8, "on", 2) == 0 || \
		  strncasecmp(line + 8, "true", 4) == 0) {
		    /*  Compact files are enabled  */
		    cmpconf = 1;
		}
	    } 
	}
	fclose(infil);
	infil = NULL;
	if(debflg != 0)
	    fprintf(stderr, "Debug: readagiconf(): Read %s\n", inam);
    } else if (debflg != 0) {
	fprintf(stderr, "Debug: readagiconf(): Can't open %s\n", inam);
    }
    inam[0] = 0;

    /*  Agiconf conf file  */
    if (debflg == 0)
	strcpy(inam, "/etc/namelookup-agi/confagi.conf");
    else
	strcpy(inam, "confagi.conf");
    if ((infil = fopen(inam, "r")) == NULL) {
	if(debflg != 0)
	    fprintf(stderr, "Debug: readagiconf(): Can't open %s\n", inam);
	return(-1);
    }

    while (fgets(line, 4096, infil)) {
	i = 0;
	/*  Cleanup  */
	while (line[i] != 0) {
	    if (line[i] == '#' || (line[i] <'!' && line[i] > 0)) {
		line[i] = 0;
		break;
	    }
	    i++;
	}
	/*
	*  Some countries share country codes, and some ISO CCs.
	*  The combination of the two seems to be unique.
	*
	*  If struct ctry is empty, just copy.
	*  If not, match with what is already in there.
	*  This way, both country code and ISO CC are matched.
	*
	*  Note: country is country code, cntrycc is ISO CC.
	*/
	if (strncmp(line, "country=", 8) == 0) {
	    i = 0;
	    while (i < ctrycnt) {
		/*  Match country code and try to match ISO CC  */
		if (strcmp(countries[i].num, line + 8) == 0 && \
		  (country.cc[0] == 0 || \
		  strcmp(countries[i].cc, country.cc) == 0)) {
		    memcpy(&country, &(countries[i]), sizeof(struct ctry));
		    break;
		}
		i++;
	    }
	} else if (strncmp(line, "cntrycc=", 8) == 0) {
	    i = 0;
	    while (i < ctrycnt) {
		/*  Match ISO CC and try to match country code  */
		if (strcmp(countries[i].cc, line + 8) == 0 && \
		  (country.num[0] == 0 || \
		  strcmp(countries[i].num, country.num) == 0)) {
		    memcpy(&country, &(countries[i]), sizeof(struct ctry));
		    break;
		}
		i++;
	    }
	} else if (strncmp(line, "intcpfx=", 8) == 0) {
	    if (country.incpfx[0] == 0) {
		fprintf(stderr, \
		  "Info: readagiconf(): International call prefix: %s\n", \
		  line + 8);
	    } else {
		if (strcmp(country.incpfx, line + 8) != 0) {
		    fprintf(stderr, "Info: readagiconf(): ");
		    fprintf(stderr, "International call prefixes differ: ");
		    fprintf(stderr, "Country nr %s: %s %s\n", \
		      country.num, country.incpfx, line + 8);
		}
	    }
	    strncpy(country.incpfx, line + 8, ANL_CTRY_INC_MAX);
	} else if (strncmp(line, "trnkpfx=", 8) == 0) {
	    if (country.trkpfx[0] == 0) {
		fprintf(stderr, \
		  "Info: readagiconf(): trunk prefix: %s\n", line + 8);
	    } else {
		if (strcmp(country.trkpfx, line + 8) != 0) {
		    fprintf(stderr, "Info: readagiconf(): ");
		    fprintf(stderr, "Trunk prefixes differ: ");
		    fprintf(stderr, "Country nr %s: %s %s\n", \
		      country.num, country.trkpfx, line + 8);
		}
	    }
	    strncpy(country.trkpfx, line + 8, ANL_CTRY_TRK_MAX);
	} else if (strncmp(line, "region1=fs", 10) == 0) {
	    regconf = 2;
	} else if (strncmp(line, "areafil=", 8) == 0) {
	    i = 0;
	    while (i < areacnt) {
		if (strcmp(areas[i].filnam, line + 8) == 0) {
		    /*  Cleanup first  */
		    j = i - areas[i].isocnt;
		    areas[j].sel = 0;
		    j++;
		    while (areas[j].isocnt != 0) {
			areas[j].sel = 0;
			j++;
			if(j > areacnt)
			    break;
		    }
		    /*  Then select this area file  */
		    areas[i].sel = 1;
		    /*  Tell to wrtagiconf() to include this info  */
		    arconf = 1;
		    break;
		}
		i++;
	    }
	}
    }

    if(debflg != 0)
	fprintf(stderr, "Debug: readagiconf(): Read %s\n", inam);

    return(0);
}

/*  Write namelookup.agi config file  */
int wrtagiconf(void)
{
    char line[4096], outnam[256];
    int i;
    FILE *outfil;

    i = 0;
    outfil = NULL;
    memset(line, 0, 4096);
    memset(outnam, 0, 256);

    if (debflg == 0)
	strcpy(outnam, "/etc/namelookup-agi/confagi.conf");
    else
	strcpy(outnam, "confagi.conf");

    if ((outfil = fopen(outnam, "w")) == NULL) {
	fprintf(stderr, "wrtagiconf(): Can't open %s\n", outnam);
	exit(1);
    }

    /*  Country  */
    if (country.name[0] != 0 && country.num[0] != 0) {
	fprintf(outfil, "country=%s\n", country.num);
	if(country.cc[0] != 0)
	    fprintf(outfil, "cntrycc=%s\n", country.cc);
	if(country.incpfx[0] != 0)
	    fprintf(outfil, "intcpfx=%s\n", country.incpfx);
	if(country.trkpfx[0] != 0)
	    fprintf(outfil, "trnkpfx=%s\n", country.trkpfx);
    }

    /*
    *  Note: weblookups on / off is set in agi-lookup.conf
    */

    /*  Regions  */
    if (regconf == 2) {
	/*  regconf is set on regions.db write  */
	fprintf(outfil, "region1=fs\n");
    }

    /*  Area code files  */
    if (arconf != 0) {
	/*
	*  1: Set by readagiconf()
	*  2: Set on areas.db write
	*/
	i = 0;
	while (i < areacnt) {
	    if(areas[i].sel != 0)
		fprintf(outfil, "areafil=%s\n", areas[i].filnam);
	    i++;
	}
    }

    if(debflg != 0)
	fprintf(stderr, "Debug: wrtagiconf(): Wrote %s\n", outnam);

    return(0);
}


int main(int argc, char **argv)
{
    int c, cont, i, mode, msdclk, msyptr, opt;
    char isostr[8], locstr[64], *term, *tty;

    enum { NONE, COUNTR, WEBLKP, BLOCKL, NUMS, AREA, REGION, QUIT };

    arconf  = 0;	/*  Zero if not done  */
    areacnt = 0;	/*  Nr of area code list files */
    ascflg  = 0;	/*  ASCII scroll progress bar when 1  */
    blkconf = 0;	/*  Zero if not done  */
    ctrycnt = 0;	/*  Nr of countries  */
    colflg  = 0;	/*  Color when 1  */
    cols    = 0;	/*  Screen width  */
    cmpconf = 0;	/*  Zero if unknown  */
    numcnt  = 0;	/*  Number of numbers  */
    numentr = 0;	/*  Number of numbers available  */
    numax   = 0;	/*  Max number of numbers  */
    c       = 0;
    cont    = 1;	/*  Continue while not 0  */
    debflg  = 0;	/*  Debug on when 1  */
    i       = 0;
    mode    = 1;	/*  1st menu option  */
    msdclk  = 0;	/*  Mouse double click  */
    msyptr  = 0;	/*  Mouse Y postion  */
    numconf = 0;	/*  Zero if not done  */
    opt     = 0;
    regconf = 0;	/*  Zero if unknown  */
    rows    = 0;	/*  Screen hight  */
    txtflg  = 0;	/*  Text output when 1  */
    utflg   = 0;	/*  1 when UTF-8  */
    webconf = 0;	/*  Zero if unknown  */
    home    = NULL;	/*  User home dir  */
    nums    = NULL;	/*  Phone numbers struct array  */
    term    = NULL;	/*  Terminal type  */
    tty     = NULL;	/*  TTY Name  */
    memset(areas, 0, sizeof(areas));		/*  List of area code files  */
    memset(grpnam, 0, 32);			/*  Group name  */
    memset(locstr, 0, 64);			/*  Locale  */
    memset(isostr, 0, 8);			/*  ISO CC  */
    memset(&country, 0, sizeof(country));	/*  Our country  */
    memset(countries, 0, sizeof(countries));	/*  List of countries  */
    memset(&msevent, 0, sizeof(msevent));	/*  Mouse event  */

    while ((opt = getopt(argc, argv, "acdg:htv")) != -1) {
	switch (opt) {
	    case 'a':
		ascflg = 1;
		break;
	    case 'c':
		colflg = 1;
		break;
	    case 'd':
		debflg = 1;
		break;
	    case 'g':
		strncpy(grpnam, optarg, 31);
		break;
	    case 'h':
		hlp();
		exit(0);
	    case 't':
		txtflg = 1;
		break;
	    case 'v':
		prtvers();
		exit(0);
	    default:
		hlp();
		exit(1);
	}
    }

    if (debflg != 0) {
	fprintf(stderr, "Debug: main(): Agiconf version: %s\n", version);
	#ifdef ANL_WITH_WIDE_NAMES
	 fprintf(stderr, "Debug: main(): With name length %d bytes\n", \
	   ANL_NAME_MAX);
	#endif
    }

    /*  Check locale  */
    strncpy(locstr, setlocale(LC_ALL, ""), 56);
    if(debflg != 0)
	fprintf(stderr, "Debug: main(): Locale: %s\n", locstr);
    if (locstr[2] == '_' && locstr[5] == '.') {
	strncpy(isostr, locstr + 3, 2);
	if(debflg != 0)
	    fprintf(stderr, "Debug: main(): ISO CC: %s\n", isostr);
    }
    i = 0;
    while (i < strlen(locstr)) {
	if (strncasecmp(locstr + i, "UTF-8", 5) == 0 || \
	  strncasecmp(locstr + i, "utf8", 4) == 0) {
	    /*  UTF-8 locale  */
	    utflg = 1;
	    if(debflg != 0)
		fprintf(stderr, "Debug: main(): Locale is UTF-8\n");
	    break;
	}
	i++;
    }

    /*  Check TTY  */
    if (ascflg == 0 && utflg != 0) {
	if ((tty = ttyname(0)) != NULL) {
	    if (strncmp(tty, "/dev/pts/", 9) != 0) {
		/*  Not a termimal emulator => Force no block chars  */
		ascflg = 1;
	    }
	}
	if (ascflg == 0 && (term = getenv("TERM")) != NULL) {
	    if (strcmp(term, "linux") == 0) {
		/*  Linux console => Force no block chars  */
		ascflg = 1;
	    }
	}
	if (ascflg != 0 && debflg != 0) {
	    fprintf(stderr, "Debug: main(): Terminal is not an emulator\n");
	}
    }

    /*  Get home directory  */
    if ((home = getenv("HOME")) == NULL) {
	home = getpwuid(getuid())->pw_dir;
    }
    if(debflg != 0)
	fprintf(stderr, "Debug: main(): Homedir: %s\n", home);

    /*  Init nums  */
    if(debflg != 0)
	fprintf(stderr, "Debug: main(): Allocating memory for nums\n");
    initnums();

    /*  Read country list file  */
    ctrycnt = readcoutries(countries);
    /*  Read area code files list file */
    areacnt = readareas(areas);

    /*
    *  Read agiconf config, if it exists.
    *  This reads agi-namelookup.conf to.
    */
    readagiconf();

    /*   Put tty in Ncurses mode   */
    initscr();
    if (colflg != 0 && has_colors() == TRUE) {
	start_color();
	init_pair(deftlt,  COLOR_WHITE,  COLOR_BLUE);
	init_pair(headtxt, COLOR_YELLOW, COLOR_BLUE);
	init_pair(selctd,  COLOR_BLACK,  COLOR_CYAN);
	init_pair(highlt,  COLOR_BLACK,  COLOR_WHITE);
	bkgd(COLOR_PAIR(deftlt));
    } else {
	colflg = 0;
    }
    keypad(stdscr, TRUE);
    noecho();
    /*   No cursor   */
    leaveok(stdscr, TRUE);
    curs_set(0);
    /*  Get the screen size  */
    getmaxyx(stdscr, rows, cols);
    if(debflg != 0)
	fprintf(stderr, "Debug: main(): Resolution: %dx%d\n", cols, rows);

    /* Get all the mouse events */
    mousemask(ALL_MOUSE_EVENTS, NULL);

    /*  Draw main menu  */
    move(0, 0);
    box(stdscr, 0, 0);
    drwmen(mainmen, 1, 1, 1);
    clrline(4, 25, cols - 1);
    if (country.num[0] != 0) {
	move(4, 25);
	printw("%s", country.num);
    }
    if (country.name[0] != 0) {
	move(4, 30);
	printw("%ls", country.name);
    }
    move(5, 30);
    /*  Default On  */
    if (webconf == 1) {
	printw("Web lookups: On");
    } else if(webconf == 2) {
	printw("Web lookups: Off");
    }
    move(5, 50);
    /*  Default Off  */
    if (cmpconf == 1) {
	printw("Compact files: On ");
    } else if(cmpconf == 2) {
	printw("Compact files: Off");
    }
    move(9, 30);
    if (regconf == 1) {
	printw("Without federal states");
    } else if (regconf == 2) {
	printw("With federal states");
    }
    refresh();

    /*  Main menu input  */
    while (c != EOF && cont != 0) {
	c = getch();
	switch (c) {
	    case KEY_DOWN:
		mode++;
		if (mode > QUIT)
		    mode = QUIT;
		break;
	    case KEY_UP:
		mode--;
		if(mode < COUNTR)
		    mode = COUNTR;
		break;
	    case KEY_PPAGE:
	    case KEY_HOME:
	    case '1':
		mode = COUNTR;
		break;
	    case '2':
		mode = WEBLKP;
		break;
	    case '3':
		mode = BLOCKL;
		break;
	    case '4':
		mode = NUMS;
		break;
	    case '5':
		mode = AREA;
		break;
	    case '6':
		mode = REGION;
		break;
	    case KEY_NPAGE:
	    case KEY_END:
	    case 'Q':
	    case 'q':
		mode = QUIT;
		break;
	    case 10:
	    case 13:
	    case KEY_ENTER:
		/*  Execute  */
		if (mode == COUNTR) {
		    selctry(isostr);
		} else if (mode == WEBLKP) {
		    genconf();
		} else if (mode == BLOCKL) {
		    convblockl();
		} else if (mode == NUMS) {
		    convnums();
		} else if (mode == AREA) {
		    selarea();
		} else if (mode == REGION) {
		    selreg();
		} else if (mode == QUIT) {
		    /*  Quit  */
		    cont = 0;
		}
		break;
	    case 27:
		/*  Quit  */
		mode = QUIT;
		cont = 0;
		break;
	    case KEY_MOUSE:
		msdclk = 0;		/*  Double click  */
		msyptr = procmouse();	/*  Y position of mouse  */
		msdclk = (msevent.bstate & BUTTON1_DOUBLE_CLICKED);
		if (msyptr == 4) {
		    mode = COUNTR;
		    if(msdclk != 0)
			selctry(isostr);
	        } else if (msyptr == 5) {
		    mode = WEBLKP;
		    if(msdclk != 0)
			genconf();
	        } else if (msyptr == 6) {
		    mode = BLOCKL;
		    if(msdclk != 0)
			 convblockl();
	        } else if (msyptr == 7) {
		    mode = NUMS;
		    if(msdclk != 0)
			convnums();
	        } else if (msyptr == 8) {
		    mode = AREA;
		    if(msdclk != 0)
			selarea();
	        } else if (msyptr == 9) {
		    mode = REGION;
		    if(msdclk != 0)
			selreg();
	        } else if (msyptr == 11) {
		    mode = QUIT;
		    if(msdclk != 0)
			cont = 0;
	        }
		break;
	    case KEY_RESIZE:
		procresiz();
		break;
	    default:
		beep();
		if (debflg != 0) {
		    fprintf(stderr, \
		      "Debug: main(): Invalid input: %04X\n", c);
		}
		break;
	}	/*  End of switch  */
	drwmen(mainmen, 1, 1, mode);
	/*  Status  */
	clrline(4, 25, cols - 1);
	if (country.num[0] != 0) {
	    move(4, 25);
	    printw("%s", country.num);
	}
	if (country.name[0] != 0) {
	    move(4, 30);
	    printw("%ls", country.name);
	}
	clrline(5, 30, cols - 1);
	move(5, 30);
	if (webconf == 1) {
	    printw("Web lookups: On ");
	} else if (webconf == 2) {
	    printw("Web lookups: Off");
	}
	move(5, 50);
	if (cmpconf == 1) {
	    printw("Compact files: On ");
	} else if (cmpconf == 2) {
	    printw("Compact files: Off");
	}
	move(6, 30);
	if(blkconf != 0)
	    printw("Done");
	move(7, 30);
	if(numconf != 0)
	    printw("Done");
	move(8, 30);
	if (arconf == 2) {
	    /*  areas.db write in this session  */
	    printw("Done");
	}
	clrline(9, 30, cols - 1);
	move(9, 30);
	if (regconf == 1) {
	    printw("Without federal states");
	} else if (regconf == 2) {
	    printw("With federal states");
	}
	refresh();
    }

    quit(1);

    /*  Write agiconf config  */
    wrtagiconf();

    return(0);
}

