#ifndef AGI_CONF_H
#define AGI_CONF_H 1

#include <stdio.h>
#include <wchar.h>
#include <wctype.h>
#include "agi-namelookup.h"

/*
*  Files and paths
*/

/*  Source files  */
#define ANL_SRC_PATH "/usr/local/share/phonebook"

/*  Additional output files  */
#define ANL_AREA_TSV "areas.tsv"
#define ANL_REGS_TSV "regions.tsv"

/*  Define blocklist name as a struct  */
struct blknam {
	char name[ANL_NAME_SIZ];
};

/*  Menus  */
struct menit {
	char *id;
	char *name;
};

/*  Colors  */
short deftlt  = 1;
short headtxt = 2;
short selctd  = 3;
short highlt  = 4;

/*  Main menu  */
struct menit mainmen[13] = {
	{ " ", "Name and location lookup AGI configuration" }, \
	{ " ", "Select item and press <Enter>" }, \
	{ " ", " " }, \
	{ "1", "Select Country" }, \
	{ "2", "Configuration" }, \
	{ "3", "Convert Blocklist" }, \
	{ "4", "Convert Phonebook" }, \
	{ "5", "Select Areas" }, \
	{ "6", "Select Regions" }, \
	{ " ", " " }, \
	{ "Q", "Quit program" }, \
	{ "", "" }, \
	{ (char *) NULL, (char *) NULL }  \
};

/*  Coutry selection bottom  */
struct menit cntrymen[4] = { \
	{ "N", "None" }, \
	{ "Q", "Quit this menu" }, \
	{ "", "" }, \
	{ (char *) NULL, (char *) NULL }  \
};

/*  Areas selection bottom  */
struct menit areamen[6] = { \
	{ "A", "All" }, \
	{ "N", "None" }, \
	{ "W", "Write to areas file" }, \
	{ "Q", "Quit this menu" }, \
	{ "", "" }, \
	{ (char *) NULL, (char *) NULL }  \
};

/*  Configuration menu  */
struct menit confmen[11] = { \
	{ " ", "Configuration" }, \
	{ " ", " " }, \
	{ "1", "Enable web lookups" }, \
	{ "2", "Disable web lookups" }, \
	{ "3", "Enable compact files" }, \
	{ "4", "Disable compact files" }, \
	{ " ", " " }, \
	{ "W", "Write to file agi-namelookup.conf" }, \
	{ "Q", "Quit this menu" }, \
	{ "", "" }, \
	{ (char *) NULL, (char *) NULL }  \
};


/*  Region menu  */
struct menit regmen[9] = { \
	{ " ", "Select region 1 version" }, \
	{ " ", " " }, \
	{ "1", "Without federal states" }, \
	{ "2", "With federal states" }, \
	{ " ", " " }, \
	{ "W", "Write to regions file" }, \
	{ "Q", "Quit this menu" }, \
	{ "", "" }, \
	{ (char *) NULL, (char *) NULL }  \
};


/*  Names and phone numbers  */
#define ANL_DFLT_SIZ 256

/*  Countries  */
#define ANL_CTRY_NAM_SIZ  48
#define ANL_CTRY_CC_SIZ   8
#define ANL_CTRY_NUM_SIZ  8
#define ANL_CTRY_INC_SIZ  8
#define ANL_CTRY_TRK_SIZ  8
#define ANL_COUNTRIES_SIZ 512

enum { ANL_CTRY_NAM_MAX = ANL_CTRY_NAM_SIZ - 1 };
enum { ANL_CTRY_CC_MAX  = ANL_CTRY_CC_SIZ  - 1 };
enum { ANL_CTRY_NUM_MAX = ANL_CTRY_NUM_SIZ - 1 };
enum { ANL_CTRY_INC_MAX = ANL_CTRY_INC_SIZ - 1 };
enum { ANL_CTRY_TRK_MAX = ANL_CTRY_TRK_SIZ - 1 };
enum { ANL_COUNTRIES_MAX = ANL_COUNTRIES_SIZ - 1 };

struct ctry {
	/*  Wide characters for country name  */
	wchar_t name[ANL_CTRY_NAM_SIZ];
	char cc[ANL_CTRY_CC_SIZ];
	char num[ANL_CTRY_NUM_SIZ];
	/*  International call prefix  */
	char incpfx[ANL_CTRY_INC_SIZ];
	/*  Trunk prefix  */
	char trkpfx[ANL_CTRY_TRK_SIZ];
};

/*  Areas  */
#define ANL_AR_FNAM_SIZ 16
#define ANL_AR_CTRY_SIZ 40
#define ANL_AR_LANG_SIZ 16
#define ANL_AR_RES_SIZ  16
#define ANL_AREAS_SIZ   2048

enum { ANL_AR_FNAM_MAX = ANL_AR_FNAM_SIZ - 1 };
enum { ANL_AR_CTRY_MAX = ANL_AR_CTRY_SIZ - 1 };
enum { ANL_AR_LANG_MAX = ANL_AR_LANG_SIZ - 1 };
enum { ANL_AR_RES_MAX  = ANL_AR_RES_SIZ  - 1 };

struct area {
	int  sel;
	int  isocnt;
	char iso[4];	/*  ISO CC length is always 2  */
	char filnam[ANL_AR_FNAM_SIZ];
	/*  These may be non-ASCII  */
	wchar_t country[ANL_AR_CTRY_SIZ];
	wchar_t lang[ANL_AR_LANG_SIZ];
	wchar_t res[ANL_AR_RES_SIZ];
	int pad;	/*  To make multiple of eight  */
};

#endif	/*  agiconf.h  */

