#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


/*
*    x-www-form-urlencoded endoder
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/


void hlp(void)
{
    fprintf(stderr, \
      "Formurlenc encodes for application/x-www-form-urlencoded.\n");
    fprintf(stderr, "Reads from stdin, writes to stdout.\n");
    fprintf(stderr, \
      "Replaces Linefeed, Carriage-Return and CrLf-pair by %%0D%%0A.\n");
    fprintf(stderr, "Replaces space by '+'.\n");
    fprintf(stderr, \
      "And all but 0-9, A-Z, a-z, '-', '.', '_' and '~' by '%%XX',\n");
    fprintf(stderr, "where 'XX' is the hex value of the character.\n\n");

    fprintf(stderr, \
      "Usage: formurlenc [-h] [-n] [-s]\n");
    fprintf(stderr, "-h: Print this Help and exit.\n");
    fprintf(stderr, "-n: Don't normalise end of line.\n");
    fprintf(stderr, "-s: Use '%%20' instead of '+' for space.\n\n");
}


int main(int argc, char **argv)
{
    int i, nf, opt, sf;
    unsigned char c, p;
    char str[4096];

    c   = 0;	/*  Current char  */
    i   = 0;
    nf  = 0;	/*  Normalisation Flag  */
    opt = 0;
    p   = 0;	/*  Previous char  */
    sf  = 0;	/*  Space Flag  */
    memset(str, 0, 4096);

    while ((opt = getopt(argc, argv, "hns")) != -1) {
	switch (opt) {
	    case 'h':
		hlp();
		exit(0);
	    case 'n':
		nf = 1;
		break;
	    case 's':
		sf = 1;
		break;
	    default:
		hlp();
		exit(1);
	}
    }

    while (fgets(str, 4096, stdin)) {
	i = 0;
	/*
	*  RFC 3986 Unreserved Characters
	*                             - .  0 1 2 3 4 5 6 7 8 9
	*    A B C D E F G H I J K L M N O P Q R S T U V W X Y Z  _
	*    a b c d e f g h i j k l m n o p q r s t u v w x y z  ~
	*  US-ASCII
	*    ! " # $ % & ' ( ) * + , - . / 0 1 2 3 4 5 6 7 8 9 : ; < = > ?
	*  @ A B C D E F G H I J K L M N O P Q R S T U V W X Y Z [ \ ] ^ _
	*  ` a b c d e f g h i j k l m n o p q r s t u v w x y z { | } ~
	*/
	while (str[i] != 0) {
	    p = c;			/*  Previous char  */
	    c = (unsigned char) str[i];	/*  Current char  */
	    if (nf == 0) {
		/*  Normalisation  */
		if (c == 10 && p != 13) {
		    /*  Missing Cr  */
		    printf("%%0D");
		} else if (p == 13 && c != 10) {
		    /*  Missing Lf  */
		    printf("%%0A");
		}
	    }
	    if (sf == 0 && c == ' ') {
		printf("+");
	    } else if ((c < '0' && c != '-' && c != '.') || \
	      (c > '9' && c < 'A') || \
	      (c > 'Z' && c < 'a' && c != '_') || \
	      (c > 'z' && c != '~')) {
		printf("%%%02X", c);
	    } else {
		printf("%c", c);
	    }
	    i++;
	}
    }
    if (nf == 0 && c == 13) {
	/*  Last char was Cr  */
	printf("%%0A");
    }

    return(0);
}

