#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>


/*
*    Lookup telephone number
*    Required external program is curl.
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/


#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wstringop-truncation"
#pragma GCC diagnostic ignored "-Wformat-truncation"


/*  Curl  */
#define ANL_CURL_COMND "curl --no-styled-output -k -m 5 -s"

/*  Find name in HTML source  */
struct curlnam {
	char prename[256];
	char pstname[256];
	char name[256];
	int atlin;
	int atcol;
	int chrcnt;
	int lincnt;
	int prelen;
	int pstlen;
	int cmptr;
	int found;
	int namptr;
	int pad;
};

/*  Instructions for curl  */
struct curlprps {
	char preget[256];
	char pstget[256];
	char prepost[256];
	char pstpost[256];
	struct curlnam nam;
};


/*  Find name in buf  */
int findname(char *buf, int len, struct curlnam *namflt, char *name)
{
    char ch, *prenam, *pstnam;
    int *atlin, *atcol, *chrcnt, i, *cmptr, *lincnt, prelen, pstlen;
    int namlen, *namptr;
    
    /*          
    *  Dit telefoonnummer is gekoppeld aan 
    *  .</span>
    */

    ch = 0;
    i = 0;
    namlen = 0;

    prenam = namflt->prename;
    pstnam = namflt->pstname;
    atlin  = &(namflt->atlin);
    atcol  = &(namflt->atcol);
    lincnt = &(namflt->lincnt);
    chrcnt = &(namflt->chrcnt);
    cmptr  = &(namflt->cmptr);
    namptr = &(namflt->namptr);
    prelen = namflt->prelen;
    pstlen = namflt->pstlen;

    while (i < len) {
	ch = buf[i];
	(*chrcnt)++;
	if (ch > 0 && ch < ' ') {
	    /*  Cntrl  */
	    if (ch == 10) {
		/*  New line  */
		(*lincnt)++;
		if (*atlin != 0) {
		    /*  Reset character counter  */
		    *chrcnt = 0;
		}
	    }
	    /*  All whitespace to space  */
	    ch = ' ';
	}
	if (namflt->found == 0) {
	    /*  Name not yet found  */
	    if (*lincnt >= *atlin && *chrcnt >= *atcol) {
		if (prenam[0] != 0) {
		    /*  Pre name based look up  */
		    if (ch == prenam[*cmptr]) {
			(*cmptr)++;
		    } else if (*cmptr != 0) {
			*cmptr = 0;
			/*  Try again  */
			if(ch == prenam[*cmptr])
			    (*cmptr)++;
		    }
		    if (*cmptr >= prelen) {
			/*  Name is next  */
			namflt->found = 1;
		    }
		} else {
		    /*  Look up by line and column alone  */
		    namflt->found = 1;
		}
	    }
	} else {
	    /*
	    *  876543210
	    *  .</span>  Length 8
	    *          ^ Ptr
	    */
	    if (*namptr != 0 || ch != ' ') {
		/*  Ignore leading spaces  */
		namflt->name[*namptr] = ch;
		(*namptr)++;
		namflt->name[*namptr] = 0;
	    }
	    if (*namptr > 254) {
		fprintf(stderr, "findname(): Name too long\n");
		namflt->found = 2;
		namflt->name[0] = 0;
		return(-1);
	    }
	    if (*namptr >= pstlen && strncmp(namflt->name + *namptr - pstlen, \
	      pstnam, pstlen) == 0) {
		/* Finished  */
		namlen = *namptr - pstlen;
		namflt->name[namlen] = 0;
		namflt->found = 2;
		strncpy(name, namflt->name, 31);
		name[31] = 0;
		break;
	    }
	}
	i++;
    }

    if (namlen != 0) {
	i = namlen - 1;
	while (i > 0) {
	    if (name[i] == ' ') {
		/*  Remove trailing space  */
		name[i] = 0;
	    } else {
		namlen = i + 1;
		break;
	    }
	    i--;
	}
    }

    return(namlen);
}

/*  Check for quotes  */
int chkquot(char *line, int ofs, int len)
{
    if ((line[ofs] == '\"' && line[len - 1] == '\"') || \
      (line[ofs] == '\'' && line[len - 1] == '\'')) {
	line[len - 1] = 0;
	return(ofs + 1);
    } else {
	return(ofs);
    }
}

/*  Read config file  */
int readcurlib(struct curlprps *site, char *line)
{
    int i, len, ofs;
    FILE *libfil;

    i = 0;
    len = 0;
    ofs = 0;
    libfil = NULL;
    line[0] = 0;

    if ((libfil = fopen("/etc/namelookup-agi/weblookup.conf", "r")) == NULL) {
	fprintf(stderr, "readcurlib() Cant open weblookup.conf\n");
	return(-1);
    }

    while (fgets(line, 4096, libfil)) {
	if (line[0] == 10 || line[0] == 13 || line[0] == '#')
	    continue;
	/*  Cleanup  */
	i = 1;
	while (line[i] != 0) {
	    if(line[i] == 10 || line[i] == 13) {
		line[i] = 0;
		len = i;
		break;
	    }
	    i++;
	}
	if (strncmp(line, "preget=", 7) == 0) {
	    /*  Check for quotes  */
	    ofs = chkquot(line, 7, len);
	    strncpy(site->preget, line + ofs, 255);
	    site->preget[255] = 0;
	} else if (strncmp(line, "pstget=", 7) == 0) {
	    ofs = chkquot(line, 7, len);
	    strncpy(site->pstget, line + ofs, 255);
	    site->pstget[255] = 0;
	} else if (strncmp(line, "prepost=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(site->prepost, line + ofs, 255);
	    site->prepost[255] = 0;
	} else if (strncmp(line, "pstpost=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(site->pstpost, line + ofs, 255);
	    site->pstpost[255] = 0;
	} else if (strncmp(line, "prename=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(site->nam.prename, line + ofs, 255);
	    site->nam.prename[255] = 0;
	} else if (strncmp(line, "pstname=", 8) == 0) {
	    ofs = chkquot(line, 8, len);
	    strncpy(site->nam.pstname, line + ofs, 255);
	    site->nam.pstname[255] = 0;
	} else if (strncmp(line, "atlin=", 6) == 0) {
	    site->nam.atlin = atoi(line + 6);
	    if(site->nam.atlin < 0)
		site->nam.atlin = 0;
	} else if (strncmp(line, "atcol=", 6) == 0) {
	    site->nam.atcol = atoi(line + 6);
	    if(site->nam.atcol < 0)
		site->nam.atcol = 0;
	}
    }

    fclose(libfil);

    site->nam.prelen = strlen(site->nam.prename);
    site->nam.pstlen = strlen(site->nam.pstname);

    return(0);
}


int main(int argc, char **argv)
{
    char buf[4096], comnd[256], line[4096], name[32];
    char *num, *preget, *pstget, *prepost, *pstpost;
    int i, inbuf, len;
    FILE *p1;
    struct curlprps site;

    i = 0;
    inbuf = 0;
    len = 0;
    num = NULL;
    p1 = NULL;
    memset(buf, 0, 4096);
    memset(comnd, 0, 256);
    memset(line, 0, 4096);
    memset(name, 0, 32);
    memset(&site, 0, sizeof(struct curlprps));

    preget  = site.preget;
    pstget  = site.pstget;
    prepost = site.prepost;
    pstpost = site.pstpost;

    if (argc != 2) {
	fprintf(stderr, "Usage: namelookup number\n");
	exit(1);
    }

    /*  Check for number; 0123456789  */
    i = 0;
    len = strlen(argv[1]);
    while (i < len) {
	if (argv[1][i] < '0' || argv[1][i] > '9') {
	    fprintf(stderr, "Not a number\n");
	    exit(1);
	}
	i++;
    }
    num = argv[1];

    /*  Read config  */
    if (readcurlib(&site, line) != 0)
	return(-1);

    /*  Command  */
    if (prepost[0] == 0) {
	/*  GET  */
	if (pstget[0] == 0) {
	    snprintf(comnd, 255, "%s '%s%s'", ANL_CURL_COMND, \
	      preget, num);
	} else {
	    snprintf(comnd, 255, "%s '%s%s%s'", ANL_CURL_COMND, \
	      preget, num, pstget);
	}
    } else {
	/*  POST  */
	if (pstpost[0] == 0) {
	    snprintf(comnd, 255, "%s -d '%s%s' '%s'", ANL_CURL_COMND, \
	      prepost, num, preget);
	} else {
	    snprintf(comnd, 255, "%s -d '%s%s%s' '%s'", ANL_CURL_COMND, \
	      prepost, num, pstpost, preget);
	}
    }

    /*  Open process  */
    p1 = popen(comnd, "r");
    if (p1 == NULL) {
	fprintf(stderr, "Can't open curl\n");
	exit(1);
    }

    /*  Read process  */
    while ((inbuf = fread(buf, 1, 4095, p1)) > 0) {
	buf[inbuf] = 0;
	if(site.nam.found < 2)
	    findname(buf, inbuf, &(site.nam), name);
    }
    if (name[0] != 0) {
	printf("%s\t%s\n", num, name);
    }

    pclose(p1);

    return(0);
}

